
/* $Id: cpd_pack.c,v 3.82 1999/11/03 00:32:31 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* System Header Files */

#include <stdio.h>
#include <math.h>


/* STV Header Files */

#include "stv.h"
#include "stvmac.h"
#include "cpd_proto.h"
#include "stv_glob.h"


/* Test Routines */

void	cpd_count_notset();
void	cpd_hurl_field();


/* CheckPoint Packing Routines */


/* Pack Up Local Elements */

int
cpd_pack_checkpoint_data( Hout, nout,
	params, nparams, particles, nparticles, fields, nfields )
CPD_HOLE *Hout;
int nout;
STV_PARAMPTR params;
int nparams;
STV_PARTICLEPTR particles;
int nparticles;
STV_FIELDPTR fields;
int nfields;
{
	STV_PARTICLEPTR PT;

	STV_PARTICLE_ID ptid;

	STV_PARAMPTR P;

	STV_FIELDPTR F;

	STV_REGION R;

	STV_VALUE ptr;

	int coords[STV_MAX_DIM];

	int i, j;
	int pnum;
	int size;
	int dim;
	int num;
	int tmp;
	int cc;

	/* Pack Checkpoint Marker */

	tmp = stvMarkerCp;

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

	/* Pack # of Parameters */

	P = params;

	num = 0;

	while ( P != NULL )
	{
		if ( P->aflag & stvCpOnly )
			num++;

		P = P->next;
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
	{
		printf( "%s: %d parameters to be checkpointed.\n",
			"cpd_pack_checkpoint_data()", num );
		fflush( stdout );
	}

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &num, stvInt, 1 );

	/* Pack Up Parameters */

	P = params;

	while ( P != NULL )
	{
		/* Pack Up Parameter if to be Checkpointed */

		if ( P->aflag & stvCpOnly )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf(
					"%s: packing parameter \"%s\" with %d elements.\n",
					"cpd_pack_checkpoint_data()", P->name, P->num );
				fflush( stdout );
			}

			/* Pack Parameter Marker */

			tmp = stvMarkerParam;

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			/* Pack Parameter Name */

			cpd_blow_hole( Hout, nout, stvCpdHoleDataStr, P->name,
				stvStr, 1 );

			/* Pack Parameter Vector Size */

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &(P->num),
				stvInt, 1 );

			/* Pack Parameter Vector Elements */

			for ( i=0 ; i < P->num ; i++ )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf(
						"%s: elem #%d: \"%s\".\"%s\" type=%s(%d).\n",
						"cpd_pack_checkpoint_data()",
						i, P->name, P->pnames[i],
						stv_data_type_str( P->types[i] & stvTypeMask ),
						P->types[i] );
					fflush( stdout );
				}

				/* Pack Element Name */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataStr,
					P->pnames[i], stvStr, 1 );

				/* Pack Element Data Type */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(P->types[i]), stvInt, 1 );

				/* Pack Element Data */

				if ( (cc = cpd_blow_hole( Hout, nout,
						stvCpdHoleDataData, P->vars[i],
						P->types[i] & stvTypeMask, 1 ))
					!= stvStatusOk )
				{
					STV_ERROR( "cpd_pack_checkpoint_data", cc,
						"Typed Pack Failed." );

					return( cc );
				}
			}

			/* Pack End of Parameter Marker */

			tmp = stvMarkerParamEnd;

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );
		}

		P = P->next;
	}

	/* Pack # of Fields */

	F = fields;

	num = 0;

	while ( F != NULL )
	{
		if ( F->aflag & stvCpOnly )
			num++;

		F = F->next;
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
	{
		printf( "%s: %d fields to be checkpointed.\n",
			"cpd_pack_checkpoint_data()", num );
		fflush( stdout );
	}

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &num, stvInt, 1 );

	/* Pack Up Fields */

	F = fields;

	while ( F != NULL )
	{
		/* Pack Up Field if to be Checkpointed */

		if ( F->aflag & stvCpOnly )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: packing field \"%s\".\n",
					"cpd_pack_checkpoint_data()", F->name );
				fflush( stdout );
			}

			/* Pack Field Marker */

			tmp = stvMarkerField;

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			/* Pack Field Name */

			cpd_blow_hole( Hout, nout, stvCpdHoleDataStr, F->name,
				stvStr, 1 );

			/* Pack Decomp Field */

			if ( F->decomp != NULL )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: decomp field dim=%d\n",
						"cpd_pack_checkpoint_data()",
						F->decomp->dataDim );
					fflush( stdout );
				}

				/* Pack Field Decomp Marker */

				tmp = stvMarkerFieldDecomp;

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
					stvInt, 1 );

				/* Pack Decomp Info (for future use?) */

				dim = F->decomp->dataDim;

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &dim,
					stvInt, 1 );

				if ( dim > 0 )
				{
					if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
					{
						printf( "%s: axisType=(",
							"cpd_pack_checkpoint_data()" );
						for ( i=0 ; i < dim ; i++ )
							printf( " %s", stv_decomp_type_str(
								F->decomp->axisType[i] ) );
						printf( " )\n" );
						printf( "%s: axisInfo=(",
							"cpd_pack_checkpoint_data()" );
						for ( i=0 ; i < dim ; i++ )
							printf( " %d", F->decomp->axisInfo[i] );
						printf( " )\n" );
						printf( "%s: nbounds=(",
							"cpd_pack_checkpoint_data()" );
						for ( i=0 ; i < dim ; i++ )
							printf( " %d", F->decomp->nbounds[i] );
						printf( " )\n" );
						fflush( stdout );
					}

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						F->decomp->axisType, stvInt, dim );

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						F->decomp->axisInfo, stvInt, dim );

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						F->decomp->nbounds, stvInt, dim );

					for ( i=0 ; i < dim ; i++ )
					{
						if ( F->decomp->nbounds[i] > 0 )
						{
							if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
							{
								printf( "%s: bounds[%d] =",
									"cpd_pack_checkpoint_data()", i );
								for ( j=0 ; j < F->decomp->nbounds[i] ;
										j++ )
									printf( " (%d,%d)",
										F->decomp->axisLBounds[i][j],
										F->decomp->axisUBounds[i][j] );
								printf( "\n" );
								fflush( stdout );
							}

							cpd_blow_hole( Hout, nout,
								stvCpdHoleDataInt,
								F->decomp->axisLBounds[i],
								stvInt, F->decomp->nbounds[i] );

							cpd_blow_hole( Hout, nout,
								stvCpdHoleDataInt,
								F->decomp->axisUBounds[i],
								stvInt, F->decomp->nbounds[i] );
						}
					}

					if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
					{
						printf( "%s: (glb,gub)=(",
							"cpd_pack_checkpoint_data()" );
						for ( i=0 ; i < dim ; i++ )
							printf( " (%d,%d)", F->decomp->glb[i],
								F->decomp->gub[i] );
						printf( " )\n" );
						fflush( stdout );
					}

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						F->decomp->glb, stvInt, dim );

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						F->decomp->gub, stvInt, dim );

					/* Pack Array Declarations */

					tmp = ( F->arrayOffset != NULL )
						? stvTrue : stvFalse;

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						&tmp, stvInt, 1 );

					if ( F->arrayOffset != NULL )
					{
						if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
						{
							printf( "%s: arrayOffset=(",
								"cpd_pack_checkpoint_data()" );
							for ( i=0 ; i < dim ; i++ )
								printf( " %d", F->arrayOffset[i] );
							printf( " )\n" );
							fflush( stdout );
						}

						cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
							F->arrayOffset, stvInt, dim );
					}

					if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
					{
						printf( "%s: storage_order=%s arrayDecl=(",
							"cpd_pack_checkpoint_data()",
							stv_storage_order_str( F->storage_order ) );
						for ( i=0 ; i < dim ; i++ )
							printf( " %d", F->arrayDecl[i] );
						printf( " )\n" );
						fflush( stdout );
					}

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						&(F->storage_order), stvInt, 1 );

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						F->arrayDecl, stvInt, dim );
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: time_scale=",
						"cpd_pack_checkpoint_data()" );
					stv_dump_time_str( stdout, F->time_scale,
						stvFalse );
					printf( " base_time=%d %s\n",
						F->time_base,
						stv_time_unit_str( F->time_units ) );
					printf( "%s: data_ready=%d type=%s\n",
						"cpd_pack_checkpoint_data()", F->data_ready,
						stv_data_type_str( F->type ) );
					fflush( stdout );
				}

				/* Pack Time Scale */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->time_scale), stvInt, 1 );

				/* Pack Base Time */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->time_base), stvInt, 1 );

				/* Pack Base Time Units */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->time_units), stvInt, 1 );

				/* Pack Field Data Ready Flag */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->data_ready), stvInt, 1 );

				/* Pack Data Type */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->type), stvInt, 1 );

				/* Pack Field Data */

				size = 1;

				for ( i=0 ; i < dim ; i++ )
					size *= F->arrayDecl[i];

				if ( (cc = cpd_blow_hole( Hout, nout,
						stvCpdHoleDataData, F->var, F->type, size ))
					!= stvStatusOk )
				{
					STV_ERROR( "cpd_pack_checkpoint_data", cc,
						"Typed Pack Failed." );

					return( cc );
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: prank=%d pshape=(",
						"cpd_pack_checkpoint_data()",
						F->decomp->prank );
					for ( i=0 ; i < F->decomp->prank ; i++ )
						printf( " %d", F->decomp->pshape[i] );
					printf( " ) paddr=(" );
					for ( i=0 ; i < F->decomp->prank ; i++ )
						printf( " %d", F->paddr[i] );
					printf( " )\n" );
					fflush( stdout );
				}

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->decomp->prank), stvInt, 1 );

				if ( F->decomp->prank > 0 )
				{
					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						F->decomp->pshape, stvInt, F->decomp->prank );

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						F->paddr, stvInt, F->decomp->prank );
				}
			}

			else if ( F->particle != NULL )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: particle field \"%s\"\n",
						"cpd_pack_checkpoint_data()",
						F->particle->name );
					fflush( stdout );
				}

				/* Pack Field Particle Marker */

				tmp = stvMarkerFieldParticle;

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
					stvInt, 1 );

				/* Pack Particle Name for Subsequent Match Up... */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataStr,
					F->particle->name, stvStr, 1 );

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: time_scale=",
						"cpd_pack_checkpoint_data()" );
					stv_dump_time_str( stdout, F->time_scale,
						stvFalse );
					printf( " base_time=%d %s\n",
						F->time_base,
						stv_time_unit_str( F->time_units ) );
					printf( "%s: data_ready=%d type=%s\n",
						"cpd_pack_checkpoint_data()", F->data_ready,
						stv_data_type_str( F->type ) );
					fflush( stdout );
				}

				/* Pack Time Scale */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->time_scale), stvInt, 1 );

				/* Pack Base Time */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->time_base), stvInt, 1 );

				/* Pack Base Time Units */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->time_units), stvInt, 1 );

				/* Pack Field Data Ready Flag */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->data_ready), stvInt, 1 );

				/* Pack Data Type */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->type), stvInt, 1 );

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: nelems=%d dim=%d\n",
						"cpd_pack_checkpoint_data()",
						F->nelems, F->particle->dataDim );
					fflush( stdout );
				}

				/* Pack Nelems */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					&(F->nelems), stvInt, 1 );

				/* Pack Dimension */

				dim = F->particle->dataDim;

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &dim,
					stvInt, 1 );

				/* Set Max Region */

				for ( i=0 ; i < dim ; i++ )
				{
					STV_REGION_MIN( R, i ) = F->particle->glb[i];

					STV_REGION_MAX( R, i ) = F->particle->gub[i];
				}

				/* Count Number of Particles */

				pnum = 1;

				(*(F->particle->get_particle))( pnum, R,
					&ptid, coords );

				while ( ptid != NULL )
				{
					pnum++;

					(*(F->particle->get_particle))( pnum, R,
						&ptid, coords );
				}

				pnum--;

				/* Reset get_particle() */

				(*(F->particle->get_particle))( -1, R, &ptid, coords );

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: npart=%d ",
						"cpd_pack_checkpoint_data()", pnum );
					fflush( stdout );
				}

				/* Pack Number of Particles */

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &pnum,
					stvInt, 1 );

				/* Pack Particle Field Data */

				pnum = 1;

				(*(F->particle->get_particle))( pnum, R,
					&ptid, coords );

				while ( ptid != NULL )
				{
					if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
					{
						printf( "." );
						fflush( stdout );
					}

					(*(F->get_pfield))( ptid, F->get_pfield_arg, &ptr );

					/* Pack Particle Data */

					if ( (cc = cpd_blow_hole( Hout, nout,
							stvCpdHoleDataData,
							ptr, F->type, F->nelems ))
						!= stvStatusOk )
					{
						STV_ERROR( "cpd_pack_checkpoint_data", cc,
							"Typed Pack Failed." );

						return( cc );
					}

					pnum++;

					(*(F->particle->get_particle))( pnum, R,
						&ptid, coords );
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( " done.\n" );
					fflush( stdout );
				}

				/* Reset get_particle() Again */

				(*(F->particle->get_particle))( -1, R, &ptid, coords );
			}

			else
			{
				/* Pack Field Particle Marker */

				tmp = stvMarkerFieldNone;

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
					stvInt, 1 );

				sprintf( STVERRTXT,
					"Warning: No Decomp/Particle for Field \"%s\"",
					F->name );

				STV_ERROR( "cpd_pack_checkpoint_data",
					stvStatusBadParam, STVERRTXT );
			}

			/* Pack End of Field Marker */

			tmp = stvMarkerFieldEnd;

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );
		}

		F = F->next;
	}

	/* Pack # of Base Particles */

	PT = particles;

	num = 0;

	while ( PT != NULL )
	{
		if ( PT->make_particle != NULL )
			num++;

		PT = PT->next;
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
	{
		printf( "%s: %d base particles to be checkpointed.\n",
			"cpd_pack_checkpoint_data()", num );
		fflush( stdout );
	}

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &num, stvInt, 1 );

	/* Pack Base Particles */

	PT = particles;

	while ( PT != NULL )
	{
		/* Pack Up Particle if to be Checkpointed */

		if ( PT->make_particle != NULL )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: base particle \"%s\" dim=%d.\n",
					"cpd_pack_checkpoint_data()",
					PT->name, PT->dataDim );
				printf( "%s: (glb,gub)=(",
					"cpd_pack_checkpoint_data()" );
				for ( i=0 ; i < PT->dataDim ; i++ )
					printf( " (%d,%d)", PT->glb[i], PT->gub[i] );
				printf( " )\n" );
				printf( "%s: ntasks=%d\n",
					"cpd_pack_checkpoint_data()", PT->ntasks );
				fflush( stdout );
			}

			/* Pack Particle Marker */

			tmp = stvMarkerParticle;

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			/* Pack Particle Name */

			cpd_blow_hole( Hout, nout, stvCpdHoleDataStr, PT->name,
				stvStr, 1 );

			/* Pack Particle Dimension */

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
				&(PT->dataDim), stvInt, 1 );

			/* Pack Global Bounds */

			if ( PT->dataDim > 0 )
			{
				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					PT->glb, stvInt, PT->dataDim );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
					PT->gub, stvInt, PT->dataDim );
			}

			/* Pack NProcs */

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &(PT->ntasks),
				stvInt, 1 );

			/* Set Max Region */

			for ( i=0 ; i < PT->dataDim ; i++ )
			{
				STV_REGION_MIN( R, i ) = PT->glb[i];

				STV_REGION_MAX( R, i ) = PT->gub[i];
			}

			/* Count Number of Particles */

			pnum = 1;

			(*(PT->get_particle))( pnum, R, &ptid, coords );

			while ( ptid != NULL )
			{
				pnum++;

				(*(PT->get_particle))( pnum, R, &ptid, coords );
			}

			pnum--;

			/* Reset get_particle() */

			(*(PT->get_particle))( -1, R, &ptid, coords );

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: npart=%d ",
					"cpd_pack_checkpoint_data()", pnum );
				fflush( stdout );
			}

			/* Pack Number of Particles */

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &pnum,
				stvInt, 1 );

			if ( PT->dataDim > 0 )
			{
				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: particle coords:",
						"cpd_pack_checkpoint_data()" );
					fflush( stdout );
				}

				/* Pack Particle Coordinates */

				pnum = 1;

				(*(PT->get_particle))( pnum, R, &ptid, coords );

				while ( ptid != NULL )
				{
					if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
					{
						printf( " (" );
						for ( i=0 ; i < PT->dataDim ; i++ )
							printf( " %d", coords[i] );
						printf( " )" );
						fflush( stdout );
					}

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
						coords, stvInt, PT->dataDim );

					pnum++;

					(*(PT->get_particle))( pnum, R, &ptid, coords );
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( " done.\n" );
					fflush( stdout );
				}

				/* Reset get_particle() Again */

				(*(PT->get_particle))( -1, R, &ptid, coords );
			}

			/* Pack End of Particle Marker */

			tmp = stvMarkerParticleEnd;

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );
		}

		PT = PT->next;
	}

	/* Pack End of Checkpoint Marker */

	tmp = stvMarkerCpEnd;

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
	{
		printf( "%s: End of Checkpoint.\n",
			"cpd_pack_checkpoint_data()" );
		fflush( stdout );
	}

	return( stvStatusOk );
}


/* Unpack Local Elements */

int
cpd_unpack_checkpoint_data( rbuf, fwdbuf, fd )
int rbuf;
int fwdbuf;
int fd;
{
	CPD_HOLE Hout[2];
	CPD_HOLE Hin;

	int nout;
	int cc;

	/* Set Up Holes */

	Hin.type = stvCpdHoleTypeUnpack;
	Hin.parm = rbuf;

	Hout[0].type = stvCpdHoleTypeWrite;
	Hout[0].parm = fd;

	if ( fwdbuf >= 0 )
	{
		Hout[1].type = stvCpdHoleTypePack;
		Hout[1].parm = fwdbuf;

		nout = 2;
	}

	else
		nout = 1;

	/* Chew Checkpoint */

	cc = cpd_chew_checkpoint_hole( &Hin, Hout, nout,
		(CPD_DATAPTR *) NULL );

	return( cc );
}


/* Read Checkpoint Data From File & Pack Up for App */

int
cpd_extract_checkpoint_data( fd, sbuf )
int fd;
int sbuf;
{
	CPD_HOLE Hout;
	CPD_HOLE Hin;

	int cc;

	/* Set Up Holes */

	Hin.type = stvCpdHoleTypeRead;
	Hin.parm = fd;

	Hout.type = stvCpdHoleTypePack;
	Hout.parm = sbuf;

	/* Chew Checkpoint */

	cc = cpd_chew_checkpoint_hole( &Hin, &Hout, 1,
		(CPD_DATAPTR *) NULL );

	return( cc );
}


/* Unpack Checkpoint Message from CP Daemon, Save in Local Structs */

int
cpd_receive_checkpoint_data( rbuf, cpdata )
int rbuf;
CPD_DATAPTR *cpdata;
{
	CPD_HOLE Hin;

	int cc;

	/* Set Up Holes */

	Hin.type = stvCpdHoleTypeUnpack;
	Hin.parm = rbuf;

	/* Chew Checkpoint */

	cc = cpd_chew_checkpoint_hole( &Hin, (CPD_HOLE *) NULL, 0, cpdata );

	return( cc );
}


/* Chew on Checkpoint Hole */

int
cpd_chew_checkpoint_hole( Hin, Hout, nout, cpdata )
CPD_HOLE *Hin;
CPD_HOLE *Hout;
int nout;
CPD_DATAPTR *cpdata;
{
	CPD_PARTICLE_COORDS C;

	CPD_PARTICLE_DATA PD;

	STV_PARTICLEPTR PT;

	STV_PARAMPTR P;

	STV_FIELDPTR F;

	CPD_DATAPTR DF;
	CPD_DATAPTR DP;
	CPD_DATAPTR D;

	STV_VALUE var;

	char name[1024];

	int ivals[STV_MAX_DIM];

	int pshape[255];
	int paddr[255];

	int *bounds;

	int save_data;
	int i, j, k;
	int nelems;
	int found;
	int npart;
	int prank;
	int pnum;
	int type;
	int size;
	int dim;
	int num;
	int tmp;
	int cc;

	save_data = stvFalse;

	if ( cpdata != NULL )
	{
		/* Initialize Ptr to Data List */

		D = *cpdata;

		if ( D != NULL )
		{
			while ( D->next != NULL )
				D = D->next;
		}

		save_data = stvTrue;
	}

	/* Unpack Checkpoint Marker */

	cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

	if ( tmp != stvMarkerCp )
	{
		sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
			"stvMarkerCp", stvMarkerCp, tmp );

		STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
			STVERRTXT );

		return( stvStatusBadParam );
	}

	/* Unpack # of Parameters */

	cpd_suck_hole( Hin, stvCpdHoleDataInt, &num, stvInt, 1 );

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &num, stvInt, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
	{
		printf( "%s: %d parameters in checkpoint.\n",
			"cpd_chew_checkpoint_hole", num );
		fflush( stdout );
	}

	/* Unpack Parameters */

	for ( i=0 ; i < num ; i++ )
	{
		/* Unpack Parameter Marker */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		if ( tmp != stvMarkerParam )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerParam", stvMarkerParam, tmp );

			STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
				STVERRTXT );

			return( stvStatusBadParam );
		}

		if ( save_data )
		{
			/* Get Next Data Struct */

			if ( D == NULL )
				D = *cpdata = cpd_create_data();

			else
				D = D->next = cpd_create_data();

			STV_NO_MEM_return( D, "cpd_chew_checkpoint_hole" );

			/* Get Param Struct */

			P = D->param = stv_create_param();

			STV_NO_MEM_return( P, "cpd_chew_checkpoint_hole" );

			P->aflag = stvCpOnly;
		}

		/* Unpack Parameter Name */

		cpd_suck_hole( Hin, stvCpdHoleDataStr, name, stvStr, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataStr, name, stvStr, 1 );

		if ( save_data )
			P->name = stv_copy_str( name );

		/* Unpack Parameter Vector Size */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &pnum, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &pnum,
			stvInt, 1 );

		if ( save_data )
		{
			P->num = pnum;

			P->pnames = (char **) MALLOC( (STV_SIZE_T) P->num
				* sizeof( char * ) );
			STV_MEMCHECK_return( P->pnames,
				"Parameter Vector Names List" );

			P->vars = (STV_VALUE *) MALLOC( (STV_SIZE_T) P->num
				* sizeof( STV_VALUE ) );
			STV_MEMCHECK_return( P->vars,
				"Parameter Vector Values List" );

			P->types = (int *) MALLOC( (STV_SIZE_T) P->num
				* sizeof( int ) );
			STV_MEMCHECK_return( P->types,
				"Parameter Vector Types List" );
		}

		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( "%s: parameter \"%s\" nelems=%d.\n",
				"cpd_chew_checkpoint_hole()", name, pnum );
			fflush( stdout );
		}

		/* Unpack Parameter Vector Elements */

		for ( j=0 ; j < pnum ; j++ )
		{
			/* Unpack Element Name */

			cpd_suck_hole( Hin, stvCpdHoleDataStr, name, stvStr, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataStr, name,
				stvStr, 1 );

			/* Unpack Element Data Type */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &type, stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &type,
				stvInt, 1 );

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: elem #%d: \"%s\" type=%s(%d).\n",
					"cpd_chew_checkpoint_hole()", j, name,
					stv_data_type_str( type & stvTypeMask ), type );
				fflush( stdout );
			}

			/* Unpack Element Data */

			var = (STV_VALUE) MALLOC(
				STV_SIZEOF( type & stvTypeMask ) );
			STV_MEMCHECK_return( var, "Parameter Value" );

			if ( (cc = cpd_suck_hole( Hin, stvCpdHoleDataData,
					var, type & stvTypeMask, 1 ))
				!= stvStatusOk )
			{
				STV_ERROR( "cpd_chew_checkpoint_hole", cc,
					"Typed Unpack Failed." );

				FREE( var );

				return( cc );
			}

			if ( (cc = cpd_blow_hole( Hout, nout, stvCpdHoleDataData,
					var, type & stvTypeMask, 1 ))
				!= stvStatusOk )
			{
				STV_ERROR( "cpd_chew_checkpoint_hole", cc,
					"Typed Pack Failed." );

				FREE( var );

				return( cc );
			}

			if ( save_data )
			{
				P->pnames[j] = stv_copy_str( name );

				P->types[j] = type;

				P->vars[j] = var;
			}

			else
				FREE( var );
		}

		/* Unpack End of Parameter Marker */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		if ( tmp != stvMarkerParamEnd )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerParamEnd", stvMarkerParamEnd, tmp );

			STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
				STVERRTXT );

			return( stvStatusBadParam );
		}
	}

	/* Unpack # of Fields */

	cpd_suck_hole( Hin, stvCpdHoleDataInt, &num, stvInt, 1 );

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &num, stvInt, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
	{
		printf( "%s: %d fields in checkpoint.\n",
			"cpd_chew_checkpoint_hole", num );
		fflush( stdout );
	}

	/* Unpack Fields */

	for ( i=0 ; i < num ; i++ )
	{
		/* Unpack Field Marker */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		if ( tmp != stvMarkerField )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerField", stvMarkerField, tmp );

			STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
				STVERRTXT );

			return( stvStatusBadParam );
		}

		if ( save_data )
		{
			/* Get Next Data Struct */

			if ( D == NULL )
				D = *cpdata = cpd_create_data();

			else
				D = D->next = cpd_create_data();

			STV_NO_MEM_return( D, "cpd_chew_checkpoint_hole" );

			/* Get Field Struct */

			F = D->field = stv_create_field();

			STV_NO_MEM_return( F, "cpd_chew_checkpoint_hole" );

			F->aflag = stvCpOnly;
		}

		/* Unpack Field Name */

		cpd_suck_hole( Hin, stvCpdHoleDataStr, name, stvStr, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataStr, name, stvStr, 1 );

		if ( save_data )
			F->name = stv_copy_str( name );

		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( "%s: field \"%s\".\n",
				"cpd_chew_checkpoint_hole()", name );
			fflush( stdout );
		}

		/* Unpack Decomp / Particle Marker */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		/* Unpack Decomp Field */

		if ( tmp == stvMarkerFieldDecomp )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: decomp field.\n",
					"cpd_chew_checkpoint_hole()" );
				fflush( stdout );
			}

			if ( save_data )
			{
				F->decomp = stv_create_decomp();

				STV_NO_MEM_return( F->decomp,
					"cpd_chew_checkpoint_hole" );
			}

			/* Unpack Data Dimension */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &dim, stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &dim,
				stvInt, 1 );

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: dim=%d.\n",
					"cpd_chew_checkpoint_hole()", dim );
				fflush( stdout );
			}

			if ( save_data )
			{
				F->decomp->dataDim = dim;

				if ( dim > 0 )
				{
					F->decomp->axisType = (int *) MALLOC(
						(STV_SIZE_T) dim * sizeof( int ) );
					STV_MEMCHECK_return( F->decomp->axisType,
						"Decomp Axis Type Array" );

					F->decomp->axisInfo = (int *) MALLOC(
						(STV_SIZE_T) dim * sizeof( int ) );
					STV_MEMCHECK_return( F->decomp->axisInfo,
						"Decomp Axis Info Array" );

					F->decomp->axisLBounds = (int **) MALLOC(
						(STV_SIZE_T) dim * sizeof( int * ) );
					STV_MEMCHECK_return( F->decomp->axisLBounds,
						"Decomp Axis Lower Bounds Array" );

					F->decomp->axisUBounds = (int **) MALLOC(
						(STV_SIZE_T) dim * sizeof( int * ) );
					STV_MEMCHECK_return( F->decomp->axisUBounds,
						"Decomp Axis Upper Bound Array" );

					F->decomp->nbounds = (int *) MALLOC(
						(STV_SIZE_T) dim * sizeof( int ) );
					STV_MEMCHECK_return( F->decomp->nbounds,
						"Decomp Axis Bounds Array Sizes" );

					F->decomp->glb = (int *) MALLOC(
						(STV_SIZE_T) dim * sizeof( int ) );
					STV_MEMCHECK_return( F->decomp->glb,
						"Decomp Global Lower Bound" );

					F->decomp->gub = (int *) MALLOC(
						(STV_SIZE_T) dim * sizeof( int ) );
					STV_MEMCHECK_return( F->decomp->gub,
						"Decomp Global Upper Bound" );

					F->arrayOffset = (int *) MALLOC(
						(STV_SIZE_T) dim * sizeof( int ) );
					STV_MEMCHECK_return( F->arrayOffset,
						"Field Array Offset" );

					F->arrayDecl = (int *) MALLOC(
						(STV_SIZE_T) dim * sizeof( int ) );
					STV_MEMCHECK_return( F->arrayDecl,
						"Field Array Decl" );
				}

				else
				{
					F->decomp->axisType = (int *) NULL;
					F->decomp->axisInfo = (int *) NULL;
					F->decomp->axisLBounds = (int **) NULL;
					F->decomp->axisUBounds = (int **) NULL;
					F->decomp->nbounds = (int *) NULL;
					F->decomp->glb = (int *) NULL;
					F->decomp->gub = (int *) NULL;
					F->arrayOffset = (int *) NULL;
					F->arrayDecl = (int *) NULL;
				}
			}

			if ( dim > 0 )
			{
				/* Axis Type */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				if ( save_data )
				{
					for ( j=0 ; j < dim ; j++ )
						F->decomp->axisType[j] = ivals[j];
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: axisType=(",
						"cpd_chew_checkpoint_hole()" );
					for ( j=0 ; j < dim ; j++ )
						printf( " %s",
							stv_decomp_type_str( ivals[j] ) );
					printf( " )\n" );
					fflush( stdout );
				}

				/* Axis Info */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				if ( save_data )
				{
					for ( j=0 ; j < dim ; j++ )
						F->decomp->axisInfo[j] = ivals[j];
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: axisInfo=(",
						"cpd_chew_checkpoint_hole()" );
					for ( j=0 ; j < dim ; j++ )
						printf( " %d", ivals[j] );
					printf( " )\n" );
					fflush( stdout );
				}

				/* Number of Coordinate Bounds Per Axis */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: nbounds=(",
						"cpd_chew_checkpoint_hole()" );
					for ( j=0 ; j < dim ; j++ )
						printf( " %d", ivals[j] );
					printf( " )\n" );
					fflush( stdout );
				}

				/* Determine Max Number of Cordinate Bounds */

				if ( save_data )
				{
					for ( j=0 ; j < dim ; j++ )
						F->decomp->nbounds[j] = ivals[j];
				}

				else
				{
					tmp = 1;

					for ( j=0 ; j < dim ; j++ )
					{
						if ( ivals[j] > tmp )
							tmp = ivals[j];
					}

					if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
					{
						printf( "%s: max nbounds = %d\n",
							"cpd_chew_checkpoint_hole()", tmp );
						fflush( stdout );
					}

					bounds = (int *) MALLOC(
						(STV_SIZE_T) tmp * sizeof( int ) );
					STV_MEMCHECK_return( bounds,
						"Temporary Decomp Axis Bounds Array" );
				}

				for ( j=0 ; j < dim ; j++ )
				{
					if ( ivals[j] > 0 )
					{
						if ( save_data )
						{
							bounds = F->decomp->axisLBounds[j] =
								(int *) MALLOC( (STV_SIZE_T) ivals[j]
									* sizeof( int ) );
							STV_MEMCHECK_return( bounds,
								"Decomp Axis Lower Bounds Array" );
						}

						cpd_suck_hole( Hin, stvCpdHoleDataInt,
							bounds, stvInt, ivals[j] );

						cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
							bounds, stvInt, ivals[j] );

						if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
						{
							printf( "%s: lower bounds = (",
								"cpd_chew_checkpoint_hole()" );
							for ( k=0 ; k < ivals[j] ; k++ )
								printf( " %d", bounds[k] );
							printf( " )\n" );
							fflush( stdout );
						}

						if ( save_data )
						{
							bounds = F->decomp->axisUBounds[j] =
								(int *) MALLOC( (STV_SIZE_T) ivals[j]
									* sizeof( int ) );
							STV_MEMCHECK_return( bounds,
								"Decomp Axis Upper Bounds Array" );
						}

						cpd_suck_hole( Hin, stvCpdHoleDataInt,
							bounds, stvInt, ivals[j] );

						cpd_blow_hole( Hout, nout, stvCpdHoleDataInt,
							bounds, stvInt, ivals[j] );

						if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
						{
							printf( "%s: upper bounds = (",
								"cpd_chew_checkpoint_hole()" );
							for ( k=0 ; k < ivals[j] ; k++ )
								printf( " %d", bounds[k] );
							printf( " )\n" );
							fflush( stdout );
						}
					}

					else if ( save_data )
					{
						F->decomp->axisLBounds[j] = (int *) NULL;
						F->decomp->axisUBounds[j] = (int *) NULL;
					}
				}

				/* GLB */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				if ( save_data )
				{
					for ( j=0 ; j < dim ; j++ )
						F->decomp->glb[j] = ivals[j];
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: glb=(",
						"cpd_chew_checkpoint_hole()" );
					for ( j=0 ; j < dim ; j++ )
						printf( " %d", ivals[j] );
					printf( " )\n" );
					fflush( stdout );
				}

				/* GUB */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				if ( save_data )
				{
					for ( j=0 ; j < dim ; j++ )
						F->decomp->gub[j] = ivals[j];
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: gub=(",
						"cpd_chew_checkpoint_hole()" );
					for ( j=0 ; j < dim ; j++ )
						printf( " %d", ivals[j] );
					printf( " )\n" );
					fflush( stdout );
				}

				/* Unpack Array Offset */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
					stvInt, 1 );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
					stvInt, 1 );

				if ( tmp )
				{
					if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
					{
						printf( "%s: arrayOffset=(",
							"cpd_chew_checkpoint_hole()" );
						fflush( stdout );
					}

					cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals,
						stvInt, dim );

					cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
						stvInt, dim );

					if ( save_data )
					{
						for ( j=0 ; j < dim ; j++ )
							F->arrayOffset[j] = ivals[j];
					}

					if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
					{
						for ( j=0 ; j < dim ; j++ )
							printf( " %d", ivals[j] );
						printf( " )\n" );
						fflush( stdout );
					}
				}

				else if ( save_data && F->arrayOffset != NULL )
				{
					FREE( F->arrayOffset );

					F->arrayOffset = (int *) NULL;
				}

				/* Unpack Storage Order */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
					stvInt, 1 );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
					stvInt, 1 );

				if ( save_data )
					F->storage_order = tmp;

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: storage_order = %s\n",
						"cpd_chew_checkpoint_hole()",
						stv_storage_order_str( tmp ) );
					fflush( stdout );
				}

				/* Unpack Array Decl */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				if ( save_data )
				{
					for ( j=0 ; j < dim ; j++ )
						F->arrayDecl[j] = ivals[j];
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: arrayDecl=(",
						"cpd_chew_checkpoint_hole()" );
					for ( j=0 ; j < dim ; j++ )
						printf( " %d", ivals[j] );
					printf( " )\n" );
					fflush( stdout );
				}
			}

			/* Unpack Time Scale */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			if ( save_data )
				F->time_scale = tmp;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: time_scale=",
					"cpd_chew_checkpoint_hole()" );
				stv_dump_time_str( stdout, tmp, stvTrue );
				fflush( stdout );
			}

			/* Unpack Base Time */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			if ( save_data )
				F->time_base = tmp;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: base_time=%d ",
					"cpd_chew_checkpoint_hole()", tmp );
				fflush( stdout );
			}

			/* Unpack Base Time Units */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			if ( save_data )
				F->time_units = tmp;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s\n", stv_time_unit_str( tmp ) );
				fflush( stdout );
			}

			/* Unpack Data Ready Flag */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			if ( save_data )
				F->data_ready = tmp;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: data_ready=%d ",
					"cpd_chew_checkpoint_hole()", tmp );
				fflush( stdout );
			}

			/* Unpack Data Type */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &type, stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &type,
				stvInt, 1 );

			if ( save_data )
				F->type = type;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "type=%s\n", stv_data_type_str( type ) );
				fflush( stdout );
			}

			/* Unpack Field Data */

			size = 1;

			for ( j=0 ; j < dim ; j++ )
				size *= ivals[j];

			var = (STV_VALUE) MALLOC( (STV_SIZE_T) size
				* STV_SIZEOF( type ) );
			STV_MEMCHECK_return( var, "Field Values" );

			if ( (cc = cpd_suck_hole( Hin, stvCpdHoleDataData,
					var, type, size ))
				!= stvStatusOk )
			{
				STV_ERROR( "cpd_chew_checkpoint_hole", cc,
					"Typed Unpack Failed." );

				FREE( var );

				return( cc );
			}

			if ( (cc = cpd_blow_hole( Hout, nout,
					stvCpdHoleDataData, var, type, size ))
				!= stvStatusOk )
			{
				STV_ERROR( "cpd_chew_checkpoint_hole", cc,
					"Typed Pack Failed." );

				FREE( var );

				return( cc );
			}

			if ( save_data )
				F->var = var;

			else
				FREE( var );

			/* Prank */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &prank, stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &prank,
				stvInt, 1 );

			if ( save_data )
				F->decomp->prank = prank;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: prank=%d\n",
					"cpd_chew_checkpoint_hole()", prank );
				fflush( stdout );
			}

			if ( prank > 0 )
			{
				/* Pshape */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, pshape,
					stvInt, prank );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, pshape,
					stvInt, prank );

				if ( save_data )
				{
					F->decomp->pshape = (int *) MALLOC(
						(STV_SIZE_T) prank * sizeof( int ) );
					STV_MEMCHECK_return( F->decomp->pshape,
						"Decomp Processor Shape Array" );

					for ( j=0 ; j < prank ; j++ )
						F->decomp->pshape[j] = pshape[j];
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: pshape=(",
						"cpd_chew_checkpoint_hole()" );
					for ( j=0 ; j < prank ; j++ )
						printf( " %d", pshape[j] );
					printf( " )\n" );
					fflush( stdout );
				}

				/* Paddr */

				cpd_suck_hole( Hin, stvCpdHoleDataInt, paddr,
					stvInt, prank );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, paddr,
					stvInt, prank );

				if ( save_data )
				{
					F->paddr = (int *) MALLOC( (STV_SIZE_T) prank
						* sizeof( int ) );
					STV_MEMCHECK_return( F->paddr,
						"Processor Address Array" );

					for ( j=0 ; j < prank ; j++ )
						F->paddr[j] = paddr[j];
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "%s: paddr=(",
						"cpd_chew_checkpoint_hole()" );
					for ( j=0 ; j < prank ; j++ )
						printf( " %d", paddr[j] );
					printf( " )\n" );
					fflush( stdout );
				}
			}

			else if ( save_data )
			{
				F->decomp->pshape = (int *) NULL;
				F->paddr = (int *) NULL;
			}
		}

		/* Unpack Particle Field */

		else if ( tmp == stvMarkerFieldParticle )
		{
			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: particle field for particle ",
					"cpd_chew_checkpoint_hole()" );
				fflush( stdout );
			}

			/* Unpack Particle Name */

			cpd_suck_hole( Hin, stvCpdHoleDataStr, name, stvStr, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataStr,
				name, stvStr, 1 );

			if ( save_data )
			{
				F->particle = stv_create_particle();

				STV_NO_MEM_return( F->particle,
					"cpd_chew_checkpoint_hole" );

				F->particle->name = stv_copy_str( name );
			}

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "\"%s\"\n", name );
				fflush( stdout );
			}

			/* Unpack Time Scale */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			if ( save_data )
				F->time_scale = tmp;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: time_scale=",
					"cpd_chew_checkpoint_hole()" );
				stv_dump_time_str( stdout, tmp, stvTrue );
				fflush( stdout );
			}

			/* Unpack Base Time */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			if ( save_data )
				F->time_base = tmp;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: base_time=%d ",
					"cpd_chew_checkpoint_hole()", tmp );
				fflush( stdout );
			}

			/* Unpack Base Time Units */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			if ( save_data )
				F->time_units = tmp;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s\n", stv_time_unit_str( tmp ) );
				fflush( stdout );
			}

			/* Unpack Data Ready Flag */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp,
				stvInt, 1 );

			if ( save_data )
				F->data_ready = tmp;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: data_ready=%d ",
					"cpd_chew_checkpoint_hole()", tmp );
				fflush( stdout );
			}

			/* Unpack Data Type */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &type, stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &type,
				stvInt, 1 );

			if ( save_data )
				F->type = type;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "type=%s\n", stv_data_type_str( type ) );
				fflush( stdout );
			}

			/* Unpack Nelems */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &nelems, stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &nelems,
				stvInt, 1 );

			if ( save_data )
				F->nelems = nelems;

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: nelems=%d ",
					"cpd_chew_checkpoint_hole()", nelems );
				fflush( stdout );
			}

			/* Unpack Data Dimension */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &dim, stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &dim,
				stvInt, 1 );

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "dim=%d ", dim );
				fflush( stdout );
			}

			/* Number of Particles */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, &npart, stvInt, 1 );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &npart,
				stvInt, 1 );

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "npart=%d ", npart );
				fflush( stdout );
			}

			/* Unpack Particle Field Values */

			for ( j=0 ; j < npart ; j++ )
			{
				if ( save_data )
				{
					/* Get New Particle Data Structure */

					if ( D->pdata == NULL )
						PD = D->pdata = cpd_create_particle_data();

					else
						PD = PD->next = cpd_create_particle_data();

					STV_NO_MEM_return( PD,
						"cpd_chew_checkpoint_hole" );
				}

				/* Unpack Field Data */

				var = (STV_VALUE) MALLOC( (STV_SIZE_T) nelems
					* STV_SIZEOF( type ) );
				STV_MEMCHECK_return( var, "Particle Field Value" );

				if ( (cc = cpd_suck_hole( Hin, stvCpdHoleDataData,
						var, type, nelems ))
					!= stvStatusOk )
				{
					STV_ERROR( "cpd_chew_checkpoint_hole", cc,
							"Typed Unpack Failed." );

					FREE( var );

					return( cc );
				}

				if ( (cc = cpd_blow_hole( Hout, nout,
						stvCpdHoleDataData, var, type, nelems ))
					!= stvStatusOk )
				{
					STV_ERROR( "cpd_chew_checkpoint_hole", cc,
							"Typed Pack Failed." );

					FREE( var );

					return( cc );
				}

				if ( save_data )
					PD->value = var;

				else
					FREE( var );

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( "." );
					fflush( stdout );
				}
			}

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( " done.\n" );
				fflush( stdout );
			}
		}

		else if ( tmp != stvMarkerFieldNone )
		{
			sprintf( STVERRTXT,
				"Parse Error: Expected Decomp/Particle Marker, Got %d",
				tmp );

			STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
				STVERRTXT );

			return( stvStatusBadParam );
		}

		/* Get End of Field Marker */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		if ( tmp != stvMarkerFieldEnd )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerFieldEnd", stvMarkerFieldEnd, tmp );

			STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
				STVERRTXT );

			return( stvStatusBadParam );
		}
	}

	/* Unpack # of Particles */

	cpd_suck_hole( Hin, stvCpdHoleDataInt, &num, stvInt, 1 );

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &num, stvInt, 1 );

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
	{
		printf( "%s: %d particles in checkpoint.\n",
			"cpd_chew_checkpoint_hole", num );
		fflush( stdout );
	}

	/* Unpack Particles */

	for ( i=0 ; i < num ; i++ )
	{
		/* Unpack Particle Marker */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		if ( tmp != stvMarkerParticle )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerParticle", stvMarkerParticle, tmp );

			STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
				STVERRTXT );

			return( stvStatusBadParam );
		}

		if ( save_data )
		{
			/* Get Next Data Struct */

			if ( D == NULL )
				D = *cpdata = cpd_create_data();

			else
				D = D->next = cpd_create_data();

			STV_NO_MEM_return( D, "cpd_chew_checkpoint_hole" );

			/* Get Particle Struct */

			PT = D->particle = stv_create_particle();

			STV_NO_MEM_return( PT, "cpd_chew_checkpoint_hole" );
		}

		/* Unpack Particle Name */

		cpd_suck_hole( Hin, stvCpdHoleDataStr, name, stvStr, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataStr, name, stvStr, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( "%s: particle \"%s\" ",
				"cpd_chew_checkpoint_hole()", name );
			fflush( stdout );
		}

		/* Unpack Data Dimension */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &dim, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &dim, stvInt, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( "dim=%d\n", dim );
			fflush( stdout );
		}

		if ( save_data )
		{
			PT->name = stv_copy_str( name );

			PT->dataDim = dim;

			if ( dim > 0 )
			{
				PT->glb = (int *) MALLOC( (STV_SIZE_T) dim
					* sizeof( int ) );
				STV_MEMCHECK_return( PT->glb,
					"Particle Global Lower Bound" );

				PT->gub = (int *) MALLOC( (STV_SIZE_T) dim
					* sizeof( int ) );
				STV_MEMCHECK_return( PT->gub,
					"Particle Global Upper Bound" );
			}

			else
			{
				PT->glb = (int *) NULL;
				PT->gub = (int *) NULL;
			}
		}

		if ( dim > 0 )
		{
			/* GLB */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals, stvInt, dim );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
				stvInt, dim );

			if ( save_data )
			{
				for ( j=0 ; j < dim ; j++ )
					PT->glb[j] = ivals[j];
			}

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: glb=(",
					"cpd_chew_checkpoint_hole()" );
				for ( j=0 ; j < dim ; j++ )
					printf( " %d", ivals[j] );
				printf( " )\n" );
				fflush( stdout );
			}

			/* GUB */

			cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals, stvInt, dim );

			cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
				stvInt, dim );

			if ( save_data )
			{
				for ( j=0 ; j < dim ; j++ )
					PT->gub[j] = ivals[j];
			}

			if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
			{
				printf( "%s: gub=(",
					"cpd_chew_checkpoint_hole()" );
				for ( j=0 ; j < dim ; j++ )
					printf( " %d", ivals[j] );
				printf( " )\n" );
				fflush( stdout );
			}
		}

		/* Nproc */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		if ( save_data )
			PT->ntasks = tmp;

		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( "%s: ntasks=%d ",
				"cpd_chew_checkpoint_hole()", tmp );
			fflush( stdout );
		}

		/* Number of Particles */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &npart, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &npart,
			stvInt, 1 );

		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( "npart=%d\n", npart );
			printf( "%s: particle coords:",
				"cpd_chew_checkpoint_hole()" );
			fflush( stdout );
		}

		/* Particle Coords */

		for ( j=0 ; j < npart ; j++ )
		{
			if ( save_data )
			{
				if ( D->pcoords == NULL )
					C = D->pcoords = cpd_create_particle_coords();

				else
					C = C->next = cpd_create_particle_coords();

				STV_NO_MEM_return( C, "cpd_chew_checkpoint_hole" );
			}

			if ( dim > 0 )
			{
				cpd_suck_hole( Hin, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, ivals,
					stvInt, dim );

				if ( save_data )
				{
					for ( k=0 ; k < dim ; k++ )
						C->coords[k] = ivals[k];
				}

				if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
				{
					printf( " (" );
					for ( k=0 ; k < dim ; k++ )
						printf( " %d", ivals[k] );
					printf( " )" );
					fflush( stdout );
				}
			}
		}

		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( " done.\n" );
			fflush( stdout );
		}

		/* Unpack End of Particle Marker */

		cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

		if ( tmp != stvMarkerParticleEnd )
		{
			sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
				"stvMarkerParticleEnd", stvMarkerParticleEnd, tmp );

			STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
				STVERRTXT );

			return( stvStatusBadParam );
		}
	}

	/* Connect Particle Fields to Actual Particles */

	if ( save_data )
	{
		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( "%s: connecting particle fields to particles... ",
				"cpd_chew_checkpoint_hole()" );
			fflush( stdout );
		}

		DF = *cpdata;

		while ( DF != NULL )
		{
			if ( DF->field != NULL && DF->field->particle != NULL )
			{
				DP = *cpdata;

				found = 0;

				while ( DP != NULL && !found )
				{
					if ( DP->particle != NULL
						&& !strcmp( DF->field->particle->name,
							DP->particle->name ) )
					{
						found++;
					}

					else
						DP = DP->next;
				}

				if ( found )
				{
					stv_free_particle( &(DF->field->particle) );

					DF->field->particle = DP->particle;
				}

				else
				{
					printf(
						"%s:  Particle %s Not Found for Field %s!\n",
						"Warning in cpd_chew_checkpoint_hole()",
						DF->field->particle->name, DF->field->name );

					stv_free_particle( &(DF->field->particle) );
				}
			}

			DF = DF->next;
		}

		if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
		{
			printf( "done.\n" );
			fflush( stdout );
		}
	}

	/* Unpack End of Checkpoint Marker */

	cpd_suck_hole( Hin, stvCpdHoleDataInt, &tmp, stvInt, 1 );

	cpd_blow_hole( Hout, nout, stvCpdHoleDataInt, &tmp, stvInt, 1 );

	if ( tmp != stvMarkerCpEnd )
	{
		sprintf( STVERRTXT, "Parse Error: Expected %s = %d, Got %d",
			"stvMarkerCpEnd", stvMarkerCpEnd, tmp );

		STV_ERROR( "cpd_chew_checkpoint_hole", stvStatusBadParam,
			STVERRTXT );

		return( stvStatusBadParam );
	}

	if ( STV_OPT_VERBOSE & STV_DEBUG_CKPT_DATA )
	{
		printf( "%s: checkpoint chew done.\n",
			"cpd_chew_checkpoint_hole()" );
		fflush( stdout );
	}

	return( stvStatusOk );
}


/* Download Checkpoint Data into Local Params & Fields */
/* Return # of items left to be downloaded */

int
cpd_download_data( cpdata, params, nparams, particles, nparticles,
	fields, nfields )
CPD_DATAPTR *cpdata;
STV_PARAMPTR params;
int nparams;
STV_PARTICLEPTR particles;
int nparticles;
STV_FIELDPTR fields;
int nfields;
{
	static CPD_DATAPTR save_parts = (CPD_DATAPTR) NULL;

	CPD_PARTICLE_COORDS Cnext;
	CPD_PARTICLE_COORDS C;

	CPD_PARTICLE_DATA PDnext;
	CPD_PARTICLE_DATA PD;

	STV_PARTICLE_ID ptid;

	STV_PARTICLEPTR PT;

	STV_PARAMPTR P;

	STV_FIELDPTR F;

	CPD_DATAPTR Dlast;
	CPD_DATAPTR Dnext;
	CPD_DATAPTR Dptr;
	CPD_DATAPTR D;

	char *pstr;

	int flag;
	int left;
	int size;
	int cc;
	int i;

	/* Snag Particles First, To Insure Fields Don't Precede... */

	D = *cpdata;

	Dlast = (CPD_DATAPTR) NULL;

	left = 0;

	while ( D != NULL )
	{
		if ( D->particle != NULL )
		{
			PT = stv_get_particle_name( D->particle->name,
				particles, nparticles );

			if ( PT != NULL )
			{
				/* Verify Valid Checkpoint Match */

				if ( !stv_match_particle( D->particle, PT ) )
				{
					printf(
				"\nCheckpoint Error: Incompatible Particle \"%s\"\n\n",
						D->particle->name );
				}

				/* Verify Checkpoint Routines Defined */

				else if ( PT->make_particle == NULL )
				{
					printf(
		"\nCheckpoint Error: No Particle make_particle() \"%s\"\n\n",
						D->particle->name );
				}

				/* Make the Particles */

				else
				{
					C = D->pcoords;

					while ( C != NULL )
					{
						/* Make a Particle */

						(*(PT->make_particle))( C->coords, &(C->ptid) );

						/* Next Particle */

						C = C->next;
					}
				}

				/* Free Checkpoint Data Struct */

				Dnext = D->next;

				if ( save_parts == NULL )
					save_parts = D;

				else
				{
					Dptr = save_parts;

					while ( Dptr->next != NULL )
						Dptr = Dptr->next;

					Dptr->next = D;
				}

				D->next = (CPD_DATAPTR) NULL;

				if ( Dlast != NULL )
					Dlast->next = Dnext;

				else
					*cpdata = Dnext;

				D = Dnext;
			}

			else
			{
				Dlast = D;

				D = D->next;

				left++;
			}
		}

		else
		{
			Dlast = D;

			D = D->next;
		}
	}

	/* Now Get Params & Fields */

	D = *cpdata;

	Dlast = (CPD_DATAPTR) NULL;

	left = 0;

	while ( D != NULL )
	{
		if ( D->param != NULL )
		{
			P = stv_get_param_name( D->param->name, params, nparams );

			if ( P != NULL )
			{
				/* Verify Valid Checkpoint Match */

				if ( !stv_match_param( D->param, P ) )
				{
					printf(
					"\nCheckpoint Error: Incompatible Param \"%s\"\n\n",
						D->param->name );
				}

				/* Copy Over Checkpoint Data */

				else
				{
					for ( i=0 ; i < D->param->num ; i++ )
					{
						if ( (cc = stv_copy_converted_data(
							D->param->vars[i],
							D->param->types[i] & stvTypeMask,
							P->vars[i], P->types[i] & stvTypeMask ))
							!= stvStatusOk )
						{
							STV_ERROR( "cpd_download_data", cc,
								"Converted Type Copy Failed." );
						}

						pstr = stv_str_of_typed_data( P->vars[i],
							P->types[i] & stvTypeMask );

						printf( "Param %s = %s\n", P->name, pstr );

						FREE( pstr );

						FREE( D->param->vars[i] );

						D->param->vars[i] = (STV_VALUE) NULL;
					}

					stv_free_param( &(D->param) );
				}

				/* Free Checkpoint Data Struct */

				Dnext = D->next;

				cpd_free_data( &D );

				if ( Dlast != NULL )
					Dlast->next = Dnext;

				else
					*cpdata = Dnext;

				D = Dnext;
			}

			else
			{
				Dlast = D;

				D = D->next;

				left++;
			}
		}

		else if ( D->field != NULL )
		{
			F = stv_get_field_name( D->field->name, fields, nfields );

			if ( F != NULL )
			{
				/* Verify Valid Checkpoint Match */

				if ( stvFalse && !stv_match_field( D->field, F ) )
				{
					printf(
					"\nCheckpoint Error: Incompatible Field \"%s\"\n\n",
						D->field->name );
				}

				/* Copy Over Checkpoint Data */

				else
				{
					/* Decomposition Field */

					if ( F->decomp != NULL )
					{
						size = 1;

						for ( i=0 ; i < F->decomp->dataDim ; i++ )
							size *= F->arrayDecl[i];

						BCOPY( F->var, D->field->var, (STV_SIZE_T) size
							* STV_SIZEOF( F->type & stvTypeMask ) );

						FREE( D->field->var );

						D->field->var = (STV_VALUE) NULL;

						stv_free_decomp( &(D->field->decomp) );
					}

					/* Particle Field */

					else if ( F->particle != NULL )
					{
						/* Verify set_pfield() is there... */

						if ( F->set_pfield == NULL )
						{
							printf(
			"\nCheckpoint Error: Particle No set_pfield() \"%s\"\n\n",
								D->field->name );
						}

						/* Go For It */

						else
						{
							/* Find Particle IDs */

							Dptr = save_parts;

							flag = stvFalse;

							while ( Dptr != NULL && !flag )
							{
								if ( !strcmp( Dptr->particle->name,
									F->particle->name ) )
								{
									flag = stvTrue;
								}

								else
									Dptr = Dptr->next;
							}

							/* Found */

							if ( flag )
							{
								/* Copy Data for Each Particle */

								C = Dptr->pcoords;

								PD = D->pdata;

								while ( PD != NULL )
								{
									/* Set User Data */

									(*(F->set_pfield))( C->ptid,
										F->set_pfield_arg, PD->value );

									/* Next Particle ID */

									C = C->next;

									/* Free Particle Data Struct */

									PDnext = PD->next;

									cpd_free_particle_data( &PD );

									PD = PDnext;
								}
							}

							else
							{
								printf(
				"\nCheckpoint Error: No Particle ID Found \"%s\"\n\n",
									D->field->name );
							}
						}
					}

					else
					{
						printf(
					"Error: No Decomp/Particle Found Field \"%s\"\n",
							F->name );
					}

					stv_free_field( &(D->field) );
				}

				/* Free Checkpoint Data Struct */

				Dnext = D->next;

				cpd_free_data( &D );

				if ( Dlast != NULL )
					Dlast->next = Dnext;

				else
					*cpdata = Dnext;

				D = Dnext;
			}

			else
			{
				Dlast = D;

				D = D->next;

				left++;
			}
		}

		else
		{
			Dlast = D;

			D = D->next;
		}
	}

	/* Clean Up Particle Stuff */

	if ( !left )
	{
		D = save_parts;

		save_parts = (CPD_DATAPTR) NULL;

		while ( D != NULL )
		{
			/* Free Up Particle Coord Stuff */

			C = D->pcoords;

			while ( C != NULL )
			{
				Cnext = C->next;

				cpd_free_particle_coords( &C );

				C = Cnext;
			}

			/* Free Up Saved Checkpoint Data Struct */

			Dnext = D->next;

			cpd_free_data( &D );

			D = Dnext;
		}
	}

	return( left );
}


void
cpd_free_data_list( cpdata )
CPD_DATAPTR *cpdata;
{
	CPD_PARTICLE_COORDS Cnext;
	CPD_PARTICLE_COORDS C;

	CPD_PARTICLE_DATA PDnext;
	CPD_PARTICLE_DATA PD;

	STV_PARTICLE_ID ptid;

	STV_PARTICLEPTR PT;

	STV_PARAMPTR P;

	STV_FIELDPTR F;

	CPD_DATAPTR Dnext;
	CPD_DATAPTR Dptr;
	CPD_DATAPTR D;

	int flag;
	int left;
	int size;
	int i;

	D = *cpdata;

	*cpdata = (CPD_DATAPTR) NULL;

	while ( D != NULL )
	{
		if ( D->particle != NULL )
		{
			stv_free_particle( &(D->particle) );

			C = D->pcoords;

			while ( C != NULL )
			{
				Cnext = C->next;

				cpd_free_particle_coords( &C );

				C = Cnext;
			}
		}

		else if ( D->param != NULL )
		{
			for ( i=0 ; i < D->param->num ; i++ )
			{
				FREE( D->param->vars[i] );

				D->param->vars[i] = (STV_VALUE) NULL;
			}

			stv_free_param( &(D->param) );
		}

		else if ( D->field != NULL )
		{
			if ( D->field->decomp != NULL )
			{
				FREE( D->field->var );

				D->field->var = (STV_VALUE) NULL;

				stv_free_decomp( &(D->field->decomp) );
			}

			/* Particle Field */

			else if ( F->particle != NULL )
			{
				PD = D->pdata;

				while ( PD != NULL )
				{
					PDnext = PD->next;

					cpd_free_particle_data( &PD );

					PD = PDnext;
				}
			}

			stv_free_field( &(D->field) );
		}

		Dnext = D->next;

		cpd_free_data( &D );

		D = Dnext;
	}
}


int
cpd_suck_hole( Hin, sucktype, ptr, type, num )
CPD_HOLE *Hin;
int sucktype;
void *ptr;
int type;
int num;
{
	char *str;

	int cc;

	switch ( Hin->type )
	{
		case stvCpdHoleTypeUnpack:
		{
			switch ( sucktype )
			{
				case stvCpdHoleDataInt:
				{
					cc = stv_unpack_int( Hin->parm,
							(int *) ptr, num, 1 );

					break;
				}

				case stvCpdHoleDataLong:
				{
					cc = stv_unpack_long( Hin->parm,
							(long *) ptr, num, 1 );

					break;
				}

				case stvCpdHoleDataULong:
				{
					cc = stv_unpack_ulong( Hin->parm,
							(unsigned long *) ptr, num, 1 );

					break;
				}

				case stvCpdHoleDataStr:
				{
					cc = stv_unpack_str( Hin->parm,
							(char *) ptr );

					break;
				}

				case stvCpdHoleDataData:
				{
					cc = stv_unpack_by_type( Hin->parm,
							ptr, num, 1, type );

					break;
				}

				default:
				{
					sprintf( STVERRTXT, "Unknown Hole Data Type = %d",
						sucktype );

					STV_ERROR( "cpd_suck_hole", stvStatusBadParam,
						STVERRTXT );

					cc = stvStatusBadParam;
				}
			}

			break;
		}

		case stvCpdHoleTypeRead:
		{
			switch ( sucktype )
			{
				case stvCpdHoleDataInt:
				{
					cc = read( Hin->parm, ptr,
						(STV_SIZE_T) num * sizeof(int) );

					break;
				}

				case stvCpdHoleDataLong:
				{
					cc = read( Hin->parm, ptr,
						(STV_SIZE_T) num * sizeof(long) );

					break;
				}

				case stvCpdHoleDataULong:
				{
					cc = read( Hin->parm, ptr,
						(STV_SIZE_T) num * sizeof(unsigned long) );

					break;
				}

				case stvCpdHoleDataStr:
				{
					str = cpd_read_str( Hin->parm );

					cc = ( str != NULL ) ? stvTrue : stvFalse;

					strcpy( (char *) ptr, str );

					if ( str != NULL )
						FREE( str );

					break;
				}

				case stvCpdHoleDataData:
				{
					cc = read( Hin->parm, ptr,
						(STV_SIZE_T) num * STV_SIZEOF( type ) );

					break;
				}

				default:
				{
					sprintf( STVERRTXT, "Unknown Hole Data Type = %d",
						sucktype );

					STV_ERROR( "cpd_suck_hole", stvStatusBadParam,
						STVERRTXT );

					cc = stvStatusBadParam;
				}
			}

			break;
		}

		default:
		{
			sprintf( STVERRTXT, "Unknown Hole Type = %d", Hin->type );

			STV_ERROR( "cpd_suck_hole", stvStatusBadParam, STVERRTXT );

			cc = stvStatusBadParam;
		}
	}

	cc = ( cc >= 0 ) ? stvStatusOk : stvStatusBadParam;

	return( cc );
}


int
cpd_blow_hole( Hout, nout, blowtype, ptr, type, num )
CPD_HOLE *Hout;
int nout;
int blowtype;
void *ptr;
int type;
int num;
{
	CPD_HOLE *HP;

	int ret;
	int cc;
	int i;

	ret = stvStatusOk;

	for ( i=0 ; i < nout ; i++ )
	{
		HP = &(Hout[i]);

		switch ( HP->type )
		{
			case stvCpdHoleTypePack:
			{
				switch ( blowtype )
				{
					case stvCpdHoleDataInt:
					{
						cc = stv_pack_int( HP->parm,
								(int *) ptr, num, 1 );

						break;
					}

					case stvCpdHoleDataLong:
					{
						cc = stv_pack_long( HP->parm,
								(long *) ptr, num, 1 );

						break;
					}

					case stvCpdHoleDataULong:
					{
						cc = stv_pack_ulong( HP->parm,
								(unsigned long *) ptr, num, 1 );

						break;
					}

					case stvCpdHoleDataStr:
					{
						cc = stv_pack_str( HP->parm,
								(char *) ptr );

						break;
					}

					case stvCpdHoleDataData:
					{
						cc = stv_pack_by_type( HP->parm,
								ptr, num, 1, type );

						break;
					}

					default:
					{
						sprintf( STVERRTXT,
							"Unknown Hole Data Type = %d", blowtype );

						STV_ERROR( "cpd_blow_hole", stvStatusBadParam,
							STVERRTXT );

						cc = stvStatusBadParam;
					}
				}

				break;
			}

			case stvCpdHoleTypeWrite:
			{
				switch ( blowtype )
				{
					case stvCpdHoleDataInt:
					{
						cc = write( HP->parm, ptr,
							(STV_SIZE_T) num * sizeof(int) );

						break;
					}

					case stvCpdHoleDataLong:
					{
						cc = write( HP->parm, ptr,
							(STV_SIZE_T) num * sizeof(long) );

						break;
					}

					case stvCpdHoleDataULong:
					{
						cc = write( HP->parm, ptr,
							(STV_SIZE_T) num * sizeof(unsigned long) );

						break;
					}

					case stvCpdHoleDataStr:
					{
						cc = cpd_write_str( HP->parm, (char *) ptr );

						break;
					}

					case stvCpdHoleDataData:
					{
						cc = write( HP->parm, ptr,
							(STV_SIZE_T) num * STV_SIZEOF( type ) );

						break;
					}

					default:
					{
						sprintf( STVERRTXT,
							"Unknown Hole Data Type = %d", blowtype );

						STV_ERROR( "cpd_blow_hole", stvStatusBadParam,
							STVERRTXT );

						cc = stvStatusBadParam;
					}
				}

				break;
			}

			default:
			{
				sprintf( STVERRTXT,
					"Unknown Hole Type = %d", HP->type );

				STV_ERROR( "cpd_blow_hole", stvStatusBadParam,
					STVERRTXT );

				cc = stvStatusBadParam;
			}
		}

		if ( cc < 0 )
			ret = stvStatusBadParam;
	}

	return( ret );
}


CPD_RECONFIG_REGION
cpd_create_reconfig_region()
{
	CPD_RECONFIG_REGION tmp;

	tmp = (CPD_RECONFIG_REGION) MALLOC(
		sizeof( struct cpd_reconfig_region_struct ) );
	STV_MEMCHECK( tmp, "Checkpoint Reconfigure Region Structure",
		return( (CPD_RECONFIG_REGION) NULL ) );

	tmp->name = (char *) NULL;

	tmp->type = -1;

	tmp->regions = (STV_REGION *) NULL;
	tmp->nregions = -1;

	tmp->cells = (STV_CELL *) NULL;
	tmp->ncells = -1;

	tmp->next = (CPD_RECONFIG_REGION) NULL;

	return( tmp );
}


void
cpd_free_reconfig_region( ptr )
CPD_RECONFIG_REGION *ptr;
{
	CPD_RECONFIG_REGION RR;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;

	RR = *ptr;

	if ( RR->name != NULL )
	{
		FREE( RR->name );

		RR->name = (char *) NULL;
	}

	RR->type = -1;

	if ( RR->regions != NULL )
	{
		for ( i=0 ; i < RR->nregions ; i++ )
			stv_init_region( RR->regions[i] );

		FREE( RR->regions );

		RR->regions = (STV_REGION *) NULL;
	}

	RR->nregions = -1;

	if ( RR->cells != NULL )
	{
		for ( i=0 ; i < RR->ncells ; i++ )
			stv_init_cell( RR->cells[i] );

		FREE( RR->cells );

		RR->cells = (STV_CELL *) NULL;
	}

	RR->ncells = -1;

	RR->next = (CPD_RECONFIG_REGION) NULL;

	FREE( RR );

	*ptr = (CPD_RECONFIG_REGION) NULL;
}


void
cpd_free_reconfig_region_list( ptr )
CPD_RECONFIG_REGION *ptr;
{
	CPD_RECONFIG_REGION RRnext;
	CPD_RECONFIG_REGION RR;

	if ( ptr == NULL || *ptr == NULL )
		return;

	RR = *ptr;

	*ptr = (CPD_RECONFIG_REGION) NULL;

	while ( RR != NULL )
	{
		RRnext = RR->next;

		cpd_free_reconfig_region( &RR );

		RR = RRnext;
	}
}


CPD_RECONFIG_DATA
cpd_create_reconfig_data()
{
	CPD_RECONFIG_DATA tmp;

	int i;

	tmp = (CPD_RECONFIG_DATA) MALLOC(
		sizeof( struct cpd_reconfig_data_struct ) );
	STV_MEMCHECK( tmp, "Checkpoint Reconfigure Data Structure",
		return( (CPD_RECONFIG_DATA) NULL ) );

	tmp->name = (char *) NULL;

	tmp->type = -1;

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
	{
		tmp->offset[i] = -1;
		tmp->dim[i] = -1;
	}

	tmp->storage_order = -1;

	tmp->gstart = -1;
	tmp->gstride = -1;
	tmp->num = -1;

	tmp->data = (STV_VALUE) NULL;

	tmp->next = (CPD_RECONFIG_DATA) NULL;

	return( tmp );
}


void
cpd_free_reconfig_data( ptr )
CPD_RECONFIG_DATA *ptr;
{
	CPD_RECONFIG_DATA RD;

	int i;

	if ( ptr == NULL || *ptr == NULL )
		return;

	RD = *ptr;

	if ( RD->name != NULL )
	{
		FREE( RD->name );

		RD->name = (char *) NULL;
	}

	RD->type = -1;

	for ( i=0 ; i < STV_MAX_DIM ; i++ )
	{
		RD->offset[i] = -1;
		RD->dim[i] = -1;
	}

	RD->storage_order = -1;

	RD->gstart = -1;
	RD->gstride = -1;
	RD->num = -1;

	if ( RD->data != NULL )
	{
		FREE( RD->data );

		RD->data = (STV_VALUE) NULL;
	}

	RD->next = (CPD_RECONFIG_DATA) NULL;

	FREE( RD );

	*ptr = (CPD_RECONFIG_DATA) NULL;
}


int
cpd_check_reconfig( app, taskp )
CPD_APPPTR app;
CPD_TASKPTR taskp;
{
	CPD_DATAPTR cpdata = (CPD_DATAPTR) NULL;

	CPD_DATAPTR Dlast;
	CPD_DATAPTR D;

	CPD_HEADER header;

	CPD_HOLE Hin;

	STV_FIELDPTR F;

	int i;

	/* SPANKME */
	/* return( stvTrue ); */

	/* First Do Easy Stuff, Check for Nprocs Mismatch */

	cpd_read_header( app->cmtfile, &header );

	if ( header.ntasks != app->ntasks )
		return( stvTrue );

	/* Get Checkpoint Data for Full Comparison */

	cpd_node_data_buf( app->cmtfile, taskp->nodeid );

	Hin.type = stvCpdHoleTypeRead;
	Hin.parm = app->cmtfile;

	if ( cpd_chew_checkpoint_hole( &Hin, (CPD_HOLE *) NULL, 0,
			&cpdata ) != stvStatusOk )
		return( stvFalse );

	/* Check Field Decomps */

	D = cpdata;

	while ( D != NULL )
	{
		if ( D->field != NULL )
		{
			F = stv_get_field_name_array( D->field->name,
				taskp->fields, taskp->nfields );

			if ( F == NULL )
			{
				cpd_free_data_list( &cpdata );
				return( stvFalse );
			}

			if ( !stv_match_field( D->field, F ) )
			{
				cpd_free_data_list( &cpdata );
				return( stvTrue );
			}
		}

		D = D->next;
	}

	cpd_free_data_list( &cpdata );

	return( stvFalse );
}


int
cpd_init_reconfig_chkpt( taskp )
CPD_TASKPTR taskp;
{
	CPD_HOLE Hout;

	STV_DECOMPPTR D;

	STV_FIELDPTR Fptr;
	STV_FIELDPTR F;

	STV_PARAMPTR Pptr;
	STV_PARAMPTR P;

	STV_VALUE ptr;

	int pindex;
	int i, j;
	int size;
	int tmp;
	int cc;

	/* Assemble Fields */

	F = (STV_FIELDPTR) NULL;

	for ( i=0 ; i < taskp->nfields ; i++ )
	{
		/* Connect Next Ptrs */

		if ( F == NULL )
			Fptr = F = taskp->fields[i];

		else
			Fptr = Fptr->next = taskp->fields[i];

		Fptr->aflag = stvCpOnly;

		if ( Fptr->decomp != NULL )
		{
			/* Fill In Dummy Field Values */

			D = Fptr->decomp;

			size = 1;

			for ( j=0 ; j < D->dataDim ; j++ )
				size *= F->arrayDecl[j];

			Fptr->var = (STV_VALUE) MALLOC( (STV_SIZE_T) size
				* STV_SIZEOF( Fptr->type ) );
			STV_MEMCHECK_return( Fptr->var, "Field Frame Data" );

			ptr = Fptr->var;
			tmp = -1;

			for ( j=0 ; j < size ; j++ )
			{
				stv_copy_converted_data( (STV_VALUE) &tmp, stvInt,
					ptr, Fptr->type );

				ptr = stv_get_typed_ptr_offset( ptr, 1, Fptr->type );
			}
		}
	}

	/* Assemble Parameters */

	P = (STV_PARAMPTR) NULL;

	for ( i=0 ; i < taskp->nparams ; i++ )
	{
		/* Connect Next Ptrs */

		if ( P == NULL )
			Pptr = P = taskp->params[i];

		else
			Pptr = Pptr->next = taskp->params[i];

		Pptr->aflag = stvCpOnly;

		/* Fill In Dummy Parameter Values */

		Pptr->vars = (STV_VALUE *) MALLOC(
			(STV_SIZE_T) Pptr->num * sizeof( STV_VALUE ) );
		STV_MEMCHECK_return( Pptr->vars,
			"Parameter Vector Values List" );

		tmp = -1;

		for ( j=0 ; j < Pptr->num ; j++ )
		{
			Pptr->vars[j] = (STV_VALUE) MALLOC(
				STV_SIZEOF( Pptr->types[j] & stvTypeMask ) );
			STV_MEMCHECK_return( Pptr->vars[j],
				"Parameter Var Ptr" );

			stv_copy_converted_data( (STV_VALUE) &tmp, stvInt,
				Pptr->vars[j], Pptr->types[j] & stvTypeMask );
		}
	}

	/* Dump Empty Data */

	cpd_node_data_buf( taskp->reconfig, 0 );

	Hout.type = stvCpdHoleTypeWrite;
	Hout.parm = taskp->reconfig;

	if ( (cc = cpd_pack_checkpoint_data( &Hout, 1,
				P, taskp->nparams, (STV_PARTICLEPTR) NULL, 0,
				F, taskp->nfields ))
			!= stvStatusOk )
		return( cc );

	return( stvStatusOk );
}


int
cpd_generate_reconfig_regions( taskp )
CPD_TASKPTR taskp;
{
	CPD_RECONFIG_REGION RR;

	STV_DECOMPPTR D;

	STV_FIELDPTR F;

	STV_REGION R;

	STV_CELL C;

	int cyclesize;
	int gstart;
	int pindex;
	int gend;
	int i, j;
	int size;
	int num;

	taskp->nrregion = 0;

	for ( i=0 ; i < taskp->nfields ; i++ )
	{
		F = taskp->fields[i];

		/* Create New Reconfig Region */

		if ( taskp->rregion == NULL )
			RR = taskp->rregion = cpd_create_reconfig_region();

		else
			RR = RR->next = cpd_create_reconfig_region();

		(taskp->nrregion)++;

		RR->name = stv_copy_str( F->name );

		RR->type = F->type;

		if ( F->decomp != NULL )
		{
			/* Determine # of Regions Needed */

			stv_init_region( R );

			stv_init_cell( C );

			D = F->decomp;

			if ( F->storage_order == stvColumnMajor )
				pindex = 0;

			else if ( F->storage_order == stvRowMajor )
				pindex = D->prank - 1;

			num = 1;

			for ( j=0 ; j < D->dataDim ; j++ )
			{
				switch ( D->axisType[j] )
				{
					case stvBlock:
					{
						gstart =
							( F->paddr[ pindex ] * D->axisInfo[j] )
							+ D->glb[j];

						gend = gstart + D->axisInfo[j] - 1;

						gend = ( gend > D->gub[j] ) ? D->gub[j] : gend;

						STV_REGION_MIN( R, j ) = gstart;
						STV_REGION_MAX( R, j ) = gend;

						if ( F->storage_order == stvColumnMajor )
							pindex++;

						else if ( F->storage_order == stvRowMajor )
							pindex--;

						break;
					}

					case stvCyclic:
					{
						/*** FUCK, SHOULDN'T THERE BE A LOOP HERE?! ***/

						size = stv_get_proc_dim_size(
							stvCyclic, D->axisInfo[j],
							D->axisLBounds[j], D->axisUBounds[j],
							D->nbounds[j], D->gub[j] - D->glb[j] + 1,
							F->paddr[ pindex ], D->pshape[ pindex ] );

						cyclesize = D->axisInfo[j];

						num *= ( size + cyclesize - 1 ) / cyclesize;

						gstart = STV_GLOBAL_CYCLE_INDEX( 0, 0,
							cyclesize, D->glb[j],
							F->paddr[ pindex ], D->pshape[ pindex ] );

						gend = gstart + cyclesize - 1;

						gend = ( gend > D->gub[j] ) ? D->gub[j] : gend;

						STV_REGION_MIN( R, j ) = gstart;
						STV_REGION_MAX( R, j ) = gend;

						if ( F->storage_order == stvColumnMajor )
							pindex++;

						else if ( F->storage_order == stvRowMajor )
							pindex--;

						break;
					}

					case stvExplicit:
					{
						/*** FUCK, SHOULDN'T THERE BE A LOOP HERE?! ***/

						STV_REGION_MIN( R, j ) = D->axisLBounds[j][0];
						STV_REGION_MAX( R, j ) = D->axisUBounds[j][0];

						if ( F->storage_order == stvColumnMajor )
							pindex++;

						else if ( F->storage_order == stvRowMajor )
							pindex--;

						break;
					}

					case stvCollapse:
					{
						STV_REGION_MIN( R, j ) = D->glb[j];
						STV_REGION_MAX( R, j ) = D->gub[j];

						break;
					}

					default:
					{
						sprintf( STVERRTXT,
							"Unknown Axis Type %d for Axis #%d in %s",
							D->axisType[j], j + 1, F->name );

						STV_ERROR( "cpd_generate_reconfig_regions",
							stvStatusBadParam, STVERRTXT );

						num = 0;
					}
				}

				STV_CELL_OF_DIM( C, j ) = 1;
			}

			if ( num > 0 )
			{
				/* Allocate Regions & Cells */

				RR->regions = (STV_REGION *) MALLOC( (STV_SIZE_T) num
					* sizeof( STV_REGION ) );
				STV_MEMCHECK_return( RR->regions,
					"Checkpoint Reconfigure Region List" );

				RR->nregions = num;

				RR->cells = (STV_CELL *) MALLOC( (STV_SIZE_T) num
					* sizeof( STV_CELL ) );
				STV_MEMCHECK_return( RR->cells,
					"Checkpoint Reconfigure Cell List" );

				RR->ncells = num;

				/* Fill 'em up... */

				j = 0;

				do
				{
					stv_copy_region( RR->regions[j], R );

					stv_copy_cell( RR->cells[j], C );

					j++;
				}
				while ( cpd_incr_reconfig_region( R, F, D ) );
			}

			else
				RR->nregions = RR->ncells = 0;
		}

		else if ( F->particle != NULL )
		{
			/* Need to pass in region desired... */
			/* Then rregion is just a single copy of that. */

			RR->nregions = RR->ncells = 0;
		}

		else
			RR->nregions = RR->ncells = 0;
	}

	return( stvStatusOk );
}


int
cpd_incr_reconfig_region( R, F, D )
STV_REGION R;
STV_FIELDPTR F;
STV_DECOMPPTR D;
{
	int cyclesize;
	int gstart;
	int pindex;
	int done;
	int gend;
	int j;

	if ( F->storage_order == stvColumnMajor )
		pindex = 0;

	else if ( F->storage_order == stvRowMajor )
		pindex = D->prank - 1;

	done = 0;
	j = 0;

	do
	{
		switch ( D->axisType[ j ] )
		{
			case stvBlock:
			case stvExplicit:
			{
				if ( F->storage_order == stvColumnMajor )
					pindex++;

				else if ( F->storage_order == stvRowMajor )
					pindex--;

				break;
			}

			case stvCyclic:
			{
				cyclesize = D->axisInfo[j];

				gstart += cyclesize * D->pshape[ pindex ];

				if ( gstart > D->gub[j] )
				{
					gstart = STV_GLOBAL_CYCLE_INDEX( 0, 0,
						cyclesize, D->glb[j],
						F->paddr[ pindex ], D->pshape[ pindex ] );
				}

				else
					done++;

				gend = gstart + cyclesize - 1;

				gend = ( gend > D->gub[j] ) ? D->gub[j] : gend;

				STV_REGION_MIN( R, j ) = gstart;
				STV_REGION_MAX( R, j ) = gend;

				if ( F->storage_order == stvColumnMajor )
					pindex++;

				else if ( F->storage_order == stvRowMajor )
					pindex--;

				break;
			}

			case stvCollapse:
				break;
		}

		j++;
	}
	while ( j < D->dataDim && !done );

	if ( !done )
		return( stvFalse );

	else
		return( stvTrue );
}


int
cpd_pack_reconfig_regions( sbuf, taskp )
int sbuf;
CPD_TASKPTR taskp;
{
	CPD_RECONFIG_REGION RR;

	int i, j, k;

	stv_pack_int( sbuf, &(taskp->nrregion), 1, 1 );

	RR = taskp->rregion;

	for ( i=0 ; i < taskp->nrregion ; i++ )
	{
		stv_pack_str( sbuf, RR->name );

		stv_pack_int( sbuf, &(RR->type), 1, 1 );

		stv_pack_int( sbuf, &(RR->nregions), 1, 1 );

		for ( j=0 ; j < RR->nregions ; j++ )
		{
			for ( k=0 ; k < STV_MAX_DIM ; k++ )
			{
				stv_pack_int( sbuf,
					&(STV_REGION_MIN( RR->regions[j], k )), 1, 1 );
				stv_pack_int( sbuf,
					&(STV_REGION_MAX( RR->regions[j], k )), 1, 1 );
			}
		}

		stv_pack_int( sbuf, &(RR->ncells), 1, 1 );

		for ( j=0 ; j < RR->ncells ; j++ )
		{
			for ( k=0 ; k < STV_MAX_DIM ; k++ )
			{
				stv_pack_int( sbuf,
					&(STV_CELL_OF_DIM( RR->cells[j], k )), 1, 1 );
			}
		}
	}

	return( stvStatusOk );
}


int
cpd_unpack_reconfig_regions( rbuf, rregion, nrregion )
int rbuf;
CPD_RECONFIG_REGION *rregion;
int *nrregion;
{
	CPD_RECONFIG_REGION RR;

	char name[1024];

	int i, j, k;

	stv_unpack_int( rbuf, nrregion, 1, 1 );

	for ( i=0 ; i < *nrregion ; i++ )
	{
		if ( *rregion == NULL )
			RR = *rregion = cpd_create_reconfig_region();

		else
			RR = RR->next = cpd_create_reconfig_region();

		stv_unpack_str( rbuf, name );

		RR->name = stv_copy_str( name );

		stv_unpack_int( rbuf, &(RR->type), 1, 1 );

		stv_unpack_int( rbuf, &(RR->nregions), 1, 1 );

		RR->regions = (STV_REGION *) MALLOC( (STV_SIZE_T) RR->nregions
			* sizeof( STV_REGION ) );
		STV_MEMCHECK_return( RR->regions,
			"Checkpoint Reconfigure Region List" );

		for ( j=0 ; j < RR->nregions ; j++ )
		{
			for ( k=0 ; k < STV_MAX_DIM ; k++ )
			{
				stv_unpack_int( rbuf,
					&(STV_REGION_MIN( RR->regions[j], k )), 1, 1 );
				stv_unpack_int( rbuf,
					&(STV_REGION_MAX( RR->regions[j], k )), 1, 1 );
			}
		}

		stv_unpack_int( rbuf, &(RR->ncells), 1, 1 );

		RR->cells = (STV_CELL *) MALLOC( (STV_SIZE_T) RR->ncells
			* sizeof( STV_CELL ) );
		STV_MEMCHECK_return( RR->cells,
			"Checkpoint Reconfigure Cell List" );

		for ( j=0 ; j < RR->ncells ; j++ )
		{
			for ( k=0 ; k < STV_MAX_DIM ; k++ )
			{
				stv_unpack_int( rbuf,
					&(STV_CELL_OF_DIM( RR->cells[j], k )), 1, 1 );
			}
		}
	}

	return( stvStatusOk );
}


int
cpd_extract_reconfig_data( fd, rregion, nrregion,
	params, nparams, rdata, nrdata )
int fd;
CPD_RECONFIG_REGION rregion;
int nrregion;
STV_PARAMPTR *params;
int *nparams;
CPD_RECONFIG_DATA *rdata;
int *nrdata;
{
	CPD_DATAPTR cpdata = (CPD_DATAPTR) NULL;

	CPD_DATAPTR Dlast;
	CPD_DATAPTR D;

	CPD_HEADER header;

	CPD_HOLE Hin;

	CPD_RECONFIG_DATA RD;

	CPD_RECONFIG_REGION RR;

	STV_FIELDPTR F;

	STV_PACK_DIM PD;

	STV_PARAMPTR P;

	int flag;
	int i, j;
	int cc;

	/* Grab Checkpoint Header */

	cpd_read_header( fd, &header );

	/* Check Each Nodeid's Checkpoint for Reconfig Data */

	P = (STV_PARAMPTR) NULL;

	*nparams = 0;
	*nrdata = 0;

	for ( i=0 ; i < header.ntasks ; i++ )
	{
		/* Get Checkpoint Data */

		cpd_node_data_buf( fd, i );

		Hin.type = stvCpdHoleTypeRead;
		Hin.parm = fd;

		if ( (cc = cpd_chew_checkpoint_hole( &Hin, (CPD_HOLE *) NULL, 0,
				&cpdata )) != stvStatusOk )
			return( cc );

		/* Check Each Field */

		D = cpdata;

		while ( D != NULL )
		{
			if ( D->field != NULL )
			{
				/* Find Reconfig Regions */

				RR = rregion;

				flag = 0;

				while ( RR != NULL && !flag )
				{
					if ( !strcmp( D->field->name, RR->name ) )
						flag++;

					else
						RR = RR->next;
				}

				/* Search Field for Regions */

				if ( flag )
				{
					for ( j=0 ; j < RR->nregions ; j++ )
					{
						PD = stv_get_SubRegion_indices( D->field,
							 RR->regions[j], RR->cells[j] );

						if ( PD != NULL )
						{
							cpd_extract_elems_SubRegion( PD, D->field,
								RR->type, rdata, nrdata );
						}

						stv_free_pack_dim_list( &PD );
					}
				}
			}

			else if ( D->param != NULL )
			{
				if ( P == NULL )
					P = *params = stv_create_param();

				else
					P = P->next = stv_create_param();

				P->aflag = stvCpOnly;

				P->name = stv_copy_str( D->param->name );

				P->num = D->param->num;

				P->pnames = (char **) MALLOC( (STV_SIZE_T) P->num
					* sizeof( char * ) );
				STV_MEMCHECK_return( P->pnames,
					"Parameter Vector Names List" );

				P->vars = (STV_VALUE *) MALLOC( (STV_SIZE_T) P->num
					* sizeof( STV_VALUE ) );
				STV_MEMCHECK_return( P->vars,
					"Parameter Vector Values List" );

				P->types = (int *) MALLOC( (STV_SIZE_T) P->num
					* sizeof( int ) );
				STV_MEMCHECK_return( P->types,
					"Parameter Vector Types List" );

				for ( j=0 ; j < P->num ; j++ )
				{
					P->pnames[j] = stv_copy_str( D->param->pnames[j] );

					P->types[j] = D->param->types[j];

					P->vars[j] = (STV_VALUE) MALLOC(
						STV_SIZEOF( P->types[j] & stvTypeMask ) );
					STV_MEMCHECK_return( P->vars[j],
						"Parameter Value" );

					if ( (cc = stv_copy_converted_data(
							D->param->vars[j],
							D->param->types[j] & stvTypeMask,
							P->vars[j], P->types[j] & stvTypeMask ))
						!= stvStatusOk )
					{
						STV_ERROR( "cpd_extract_reconfig_data", cc,
							"Converted Type Copy Failed." );

						return( cc );
					}
				}
			}

			D = D->next;
		}

		cpd_free_data_list( &cpdata );
	}

	return( stvStatusOk );
}


int
cpd_extract_elems_SubRegion( PD, F, type, rdata, nrdata )
STV_PACK_DIM PD;
STV_FIELDPTR F;
int type;
CPD_RECONFIG_DATA *rdata;
int *nrdata;
{
	CPD_RECONFIG_DATA RD;

	STV_PACK_INDEX PX;

	STV_PACK_DIM PDptr;

	STV_VALUE dptr;
	STV_VALUE fptr;

	int offset;
	int num;
	int tmp;
	int cc;
	int i;

	/* Initialize Indices */

	stv_init_pack_indices( PD );

	/* Iterate & Extract Stuff */

	RD = *rdata;

	while ( RD != NULL && RD->next != NULL )
		RD = RD->next;

	do
	{
		/* Get New Reconfig Data Struct */

		if ( *rdata == NULL )
			RD = *rdata = cpd_create_reconfig_data();

		else
			RD = RD->next = cpd_create_reconfig_data();

		RD->name = stv_copy_str( F->name );

		RD->type = type;

		RD->storage_order = F->storage_order;

		/* Save Offset Index */
		/* (see stv_pack_offset_index()) */

		PDptr = PD;

		RD->offset[0] = PDptr->gindex;
		RD->dim[0] = PDptr->dim;

		PDptr = PDptr->next;
		i = 1;

		while ( PDptr != NULL )
		{
			RD->offset[i] = PDptr->gindex;
			RD->dim[i] = PDptr->dim;

			i++;

			PDptr = PDptr->next;
		}

		/* Get Iterative Index Info */

		PX = PD->current;

		/* Calculate Number of Elements */

		num = (int) ceil( (double) ( PX->max - PX->lstart )
			/ (double) PX->lstride );

		/* Save Unpack Info */

		RD->gstart = PX->gstart;
		RD->gstride = PX->gstride;
		RD->num = num;

		/* Pack Data */

		offset = stv_get_pack_offset( PD, F );

		fptr = stv_get_typed_ptr_offset( F->var, offset + PX->lstart,
			F->type );

		if ( fptr == NULL )
		{
			STV_ERROR( "cpd_extract_elems_SubRegion", stvStatusBadParam,
				"stv_get_typed_ptr_offset() failed." );

			return( stvStatusBadParam );
		}

		/* Allocate Data Storage */

		RD->data = (STV_VALUE) MALLOC( (STV_SIZE_T) num
			 * STV_SIZEOF( RD->type ) );
		STV_MEMCHECK_return( RD->data, "Reconfigure Data" );

		dptr = RD->data;

		for ( i=0 ; i < RD->num ; i++ )
		{
			stv_copy_converted_data( fptr, F->type, dptr, RD->type );

			fptr = stv_get_typed_ptr_offset( fptr, PX->lstride,
				F->type );

			dptr = stv_get_typed_ptr_offset( dptr, 1, RD->type );
		}

		(*nrdata)++;
	}
	while ( stv_incr_pack_indices( PD ) );

	/* A.O.K. */

	return( stvStatusOk );
}


int
cpd_pack_reconfig_data( sbuf, params, nparams, rdata, nrdata )
int sbuf;
STV_PARAMPTR params;
int nparams;
CPD_RECONFIG_DATA rdata;
int nrdata;
{
	CPD_RECONFIG_DATA RD;

	STV_PARAMPTR P;

	int cc;
	int i;

	/* Pack Parameter Values */

	stv_pack_int( sbuf, &nparams, 1, 1 );

	P = params;

	while ( P != NULL )
	{
		stv_pack_str( sbuf, P->name );

		stv_pack_int( sbuf, &(P->num), 1, 1 );

		for ( i=0 ; i < P->num ; i++ )
		{
			stv_pack_str( sbuf, P->pnames[i] );

			stv_pack_int( sbuf, &(P->types[i]), 1, 1 );

			if ( (cc = stv_pack_by_type( sbuf, P->vars[i], 1, 1,
				P->types[i] & stvTypeMask )) != stvStatusOk )
			{
				STV_ERROR( "cpd_pack_reconfig_data", cc,
					"Typed Pack Failed." );

				return( cc );
			}
		}

		P = P->next;
	}

	/* Pack Reconfig Field Values */

	stv_pack_int( sbuf, &nrdata, 1, 1 );

	RD = rdata;

	for ( i=0 ; i < nrdata ; i++ )
	{
		stv_pack_str( sbuf, RD->name );

		stv_pack_int( sbuf, &(RD->type), 1, 1 );

		stv_pack_int( sbuf, RD->offset, STV_MAX_DIM, 1 );
		stv_pack_int( sbuf, RD->dim, STV_MAX_DIM, 1 );

		stv_pack_int( sbuf, &(RD->storage_order), 1, 1 );

		stv_pack_int( sbuf, &(RD->gstart), 1, 1 );
		stv_pack_int( sbuf, &(RD->gstride), 1, 1 );
		stv_pack_int( sbuf, &(RD->num), 1, 1 );

		if ( (cc = stv_pack_by_type( sbuf, RD->data, RD->num, 1,
			RD->type )) != stvStatusOk )
		{
			STV_ERROR( "stv_pack_elems_SubRegion", cc,
				"Typed Pack Failed." );

			return( cc );
		}

		RD = RD->next;
	}

	return( stvStatusOk );
}


int
cpd_unpack_reconfig_data( rbuf, params, nparams, rdata, nrdata )
int rbuf;
STV_PARAMPTR *params;
int *nparams;
CPD_RECONFIG_DATA *rdata;
int *nrdata;
{
	CPD_RECONFIG_DATA RD;

	STV_PARAMPTR P;

	char name[1024];

	int i, j;
	int cc;

	/* Unpack Parameter Values */

	stv_unpack_int( rbuf, nparams, 1, 1 );

	P = (STV_PARAMPTR) NULL;

	for ( i=0 ; i < *nparams ; i++ )
	{
		if ( P == NULL )
			P = *params = stv_create_param();

		else
			P = P->next = stv_create_param();

		P->aflag = stvCpOnly;

		stv_unpack_str( rbuf, name );

		P->name = stv_copy_str( name );

		stv_unpack_int( rbuf, &(P->num), 1, 1 );

		P->pnames = (char **) MALLOC( (STV_SIZE_T) P->num
			* sizeof( char * ) );
		STV_MEMCHECK_return( P->pnames,
			"Parameter Vector Names List" );

		P->vars = (STV_VALUE *) MALLOC( (STV_SIZE_T) P->num
			* sizeof( STV_VALUE ) );
		STV_MEMCHECK_return( P->vars,
			"Parameter Vector Values List" );

		P->types = (int *) MALLOC( (STV_SIZE_T) P->num
			* sizeof( int ) );
		STV_MEMCHECK_return( P->types,
			"Parameter Vector Types List" );

		for ( j=0 ; j < P->num ; j++ )
		{
			stv_unpack_str( rbuf, name );

			P->pnames[j] = stv_copy_str( name );

			stv_unpack_int( rbuf, &(P->types[j]), 1, 1 );

			P->vars[j] = (STV_VALUE) MALLOC(
				STV_SIZEOF( P->types[j] & stvTypeMask ) );
			STV_MEMCHECK( P->vars[j], "Parameter Value",
				return( stvFalse ) );

			if ( (cc = stv_unpack_by_type( rbuf, P->vars[j], 1, 1,
				P->types[j] & stvTypeMask )) != stvStatusOk )
			{
				STV_ERROR( "cpd_unpack_reconfig_data", cc,
					"Typed Pack Failed." );

				return( cc );
			}
		}
	}

	/* Unpack Reconfig Field Values */

	stv_unpack_int( rbuf, nrdata, 1, 1 );

	for ( i=0 ; i < *nrdata ; i++ )
	{
		if ( *rdata == NULL )
			RD = *rdata = cpd_create_reconfig_data();

		else
			RD = RD->next = cpd_create_reconfig_data();

		stv_unpack_str( rbuf, name );

		RD->name = stv_copy_str( name );

		stv_unpack_int( rbuf, &(RD->type), 1, 1 );

		stv_unpack_int( rbuf, RD->offset, STV_MAX_DIM, 1 );
		stv_unpack_int( rbuf, RD->dim, STV_MAX_DIM, 1 );

		stv_unpack_int( rbuf, &(RD->storage_order), 1, 1 );

		stv_unpack_int( rbuf, &(RD->gstart), 1, 1 );
		stv_unpack_int( rbuf, &(RD->gstride), 1, 1 );
		stv_unpack_int( rbuf, &(RD->num), 1, 1 );

		/* Allocate Data Storage */

		RD->data = (STV_VALUE) MALLOC( (STV_SIZE_T) RD->num
			 * STV_SIZEOF( RD->type ) );
		STV_MEMCHECK_return( RD->data, "Reconfigure Data" );

		if ( (cc = stv_unpack_by_type( rbuf, RD->data, RD->num, 1,
			RD->type )) != stvStatusOk )
		{
			STV_ERROR( "stv_pack_elems_SubRegion", cc,
				"Typed Pack Failed." );

			return( cc );
		}

		RD = RD->next;
	}

	return( stvStatusOk );
}


int
cpd_integrate_reconfig_data( taskp, params, nparams, rdata, nrdata )
CPD_TASKPTR taskp;
STV_PARAMPTR params;
int nparams;
CPD_RECONFIG_DATA rdata;
int nrdata;
{
	CPD_DATAPTR cpdata = (CPD_DATAPTR) NULL;

	CPD_DATAPTR D;

	CPD_HEADER header;

	CPD_HOLE Hout;
	CPD_HOLE Hin;

	CPD_RECONFIG_DATA RD;

	STV_FIELDPTR Fptr;
	STV_FIELDPTR F;
	int nf;

	STV_PARAMPTR Pptr;
	STV_PARAMPTR P;
	int np;

	STV_PARTICLEPTR PTptr;
	STV_PARTICLEPTR PT;
	int npt;

	int flag;
	int i, j;
	int cc;

	/* Get Existing Checkpoint */

	cpd_node_data_buf( taskp->reconfig, 0 );

	Hin.type = stvCpdHoleTypeRead;
	Hin.parm = taskp->reconfig;

	if ( (cc = cpd_chew_checkpoint_hole( &Hin, (CPD_HOLE *) NULL, 0,
			&cpdata )) != stvStatusOk )
		return( cc );

	/* Process Parameter Data */

	if ( taskp->nsetparams < taskp->nparams )
	{
		P = params;

		while ( P != NULL )
		{
			/* Verify Param Needs to be Set */

			flag = 0;

			for ( i=0 ; i < taskp->nparams && !flag ; i++ )
			{
				if ( !strcmp( (taskp->params[i])->name, P->name ) )
				{
					cc = i;
					flag++;
				}
			}

			if ( flag && !(taskp->setparams[cc]) )
			{
				/* Set Param Value */

				D = cpdata;
				flag = 0;

				while ( D != NULL && !flag )
				{
					if ( D->param != NULL
						&& !strcmp( D->param->name, P->name ) )
					{
						flag++;
					}

					else
						D = D->next;
				}

				/* Spank in Data */

				if ( flag )
				{
					for ( j=0 ; j < P->num ; j++ )
					{
						if ( D->param->vars[j] != NULL )
							FREE( D->param->vars[j] );

						D->param->vars[j] = P->vars[j];

						P->vars[j] = (STV_VALUE) NULL;
					}

					taskp->setparams[cc] = 1;

					(taskp->nsetparams)++;
				}
			}

			P = P->next;
		}
	}

	/* Process Reconfig Data */

	RD = rdata;

	while ( RD != NULL )
	{
		/* Find Checkpoint Field */

		D = cpdata;

		flag = 0;

		while ( D != NULL && !flag )
		{
			if ( D->field != NULL
				&& !strcmp( D->field->name, RD->name ) )
			{
				flag++;
			}

			else
				D = D->next;
		}

		/* Spank in Data */

		if ( flag )
			cpd_spank_reconfig_data_into_field( D->field, RD );

		RD = RD->next;
	}

	/* Link Up Checkpoint Fields, Params & Particles */

	F = (STV_FIELDPTR) NULL;
	P = (STV_PARAMPTR) NULL;
	PT = (STV_PARTICLEPTR) NULL;

	nf = 0;
	np = 0;
	npt = 0;

	D = cpdata;

	while ( D != NULL )
	{
		if ( D->field != NULL )
		{
			cpd_hurl_field( D->field );

			if ( F == NULL )
				Fptr = F = D->field;

			else
				Fptr = Fptr->next = D->field;

			Fptr->aflag = stvCpOnly;

			nf++;
		}

		else if ( D->param != NULL )
		{
			if ( P == NULL )
				Pptr = P = D->param;

			else
				Pptr = Pptr->next = D->param;

			Pptr->aflag = stvCpOnly;

			np++;
		}

		else if ( D->particle != NULL )
		{
			if ( PT == NULL )
				PTptr = PT = D->particle;

			else
				PTptr = PTptr->next = D->particle;

			npt++;
		}

		D = D->next;
	}

	/* Re-Write Checkpoint Data Back to File */

	cpd_node_data_buf( taskp->reconfig, 0 );

	Hout.type = stvCpdHoleTypeWrite;
	Hout.parm = taskp->reconfig;

	if ( (cc = cpd_pack_checkpoint_data( &Hout,
			1, P, np, PT, npt, F, nf ))
			!= stvStatusOk )
		return( cc );

	cpd_free_data_list( &cpdata );

	return( stvStatusOk );
}


void
cpd_hurl_field( F )
STV_FIELDPTR F;
{
	STV_REGION R;

	STV_CELL C;

	int doit;

	doit = stvFalse;

	if ( doit )
	{
		stv_init_region( R );

		stv_init_cell( C );

		stv_dump_field_array( F->var, F->arrayDecl,
			F->decomp->dataDim, R, C,
			F->type, F->storage_order );
	}

	cpd_count_notset( F, R, C );
}


void
cpd_count_notset( F, R, C )
STV_FIELDPTR F;
STV_REGION R;
STV_CELL C;
{
	STV_VALUE ptr;

	char *str;

	int indices[STV_MAX_DIM];

	int notset = 0;
	int index;
	int i;

	/* Initialize Local Index */

	for ( i=0 ; i < F->decomp->dataDim ; i++ )
		indices[i] = 0;

	ptr = F->var;

	do
	{
		str = stv_str_of_typed_data( ptr, F->type );

		if ( !strcmp( str, "-1" ) )
		{
			/* printf( "[%2d,%2d,%2d]", */
				/* indices[0], indices[1], indices[2] ); */

			notset++;
		}

		FREE( str );

		ptr = stv_get_typed_ptr_offset( ptr, 1, F->type );
	}
	while ( ptr != NULL
		&& stv_incr_field_index( F->decomp->dataDim, indices,
			F->arrayDecl, R, C, F->storage_order, stvFalse ) );

	printf( "%s: notset=%d\n", F->name, notset );
}


int
cpd_spank_reconfig_data_into_field( F, RD )
STV_FIELDPTR F;
CPD_RECONFIG_DATA RD;
{
	STV_DECOMPPTR D;

	STV_VALUE fptr;
	STV_VALUE rptr;

	int gindices[STV_MAX_DIM];
	int indices[STV_MAX_DIM];

	int offset;
	int cc;
	int i;

	/* Get Decomp */

	D = F->decomp;

	/* Handle First Index */

	gindices[0] = RD->gstart;

	indices[0] = cpd_global_to_local_index( F, D, gindices[0],
		RD->dim[0] );

	/* Handle Rest of Indices */

	for ( i=1 ; i < D->dataDim ; i++ )
	{
		gindices[i] = RD->offset[i];

		indices[i] = cpd_global_to_local_index( F, D, gindices[i],
			RD->dim[i] );
	}

	/* Copy Data */

	rptr = RD->data;

	for ( i=0 ; i < RD->num ; i++ )
	{
		offset = cpd_get_reconfig_offset( RD, F, D, indices );

		fptr = stv_get_typed_ptr_offset( F->var, offset, F->type );

		if ( (cc = stv_copy_converted_data( rptr, RD->type,
			fptr, F->type )) != stvStatusOk )
		{
			STV_ERROR( "cpd_spank_reconfig_data_into_field", cc,
				"Converted Type Copy Failed." );

			return( cc );
		}

		/* Bump Reconfig Data */

		gindices[0] += RD->gstride;

		indices[0] = cpd_global_to_local_index( F, D, gindices[0],
			RD->dim[0] );

		rptr = stv_get_typed_ptr_offset( rptr, 1, RD->type );
	}

	return( stvStatusOk );
}


int
cpd_get_reconfig_offset( RD, F, D, indices )
CPD_RECONFIG_DATA RD;
STV_FIELDPTR F;
STV_DECOMPPTR D;
int *indices;
{
	int b, e, s;
	int offset;
	int index;
	int size;
	int i;

	/* Determine Compatibility of Index Order */

	if ( F->storage_order == stvColumnMajor )
	{
		b = 0;
		e = D->dataDim;
		s = 1;
	}

	else if ( F->storage_order == stvRowMajor )
	{
		b = D->dataDim - 1;
		e = -1;
		s = -1;
	}

	/* Init Offset & Size */

	offset = 0;
	size = 1;

	/* Skip First Iterative Index */

	for ( i=b ; i != e ; i += s )
	{
		if ( RD->storage_order == stvRowMajor )
			index = D->dataDim - i - 1;

		else
			index = i;

		if ( F->arrayOffset != NULL )
			offset += ( indices[ index ] + F->arrayOffset[i] ) * size;

		else
			offset += indices[ index ] * size;

		size *= F->arrayDecl[i];
	}

	return( offset );
}


int
cpd_global_to_local_index( F, D, gindex, dindex )
STV_FIELDPTR F;
STV_DECOMPPTR D;
int gindex;
int dindex;
{
	int cyclesize;
	int pindex;
	int index;
	int size;
	int i;

	/* Calculate Processor Index */

	pindex = 0;

	for ( i=0 ; i < dindex ; i++ )
	{
		switch ( D->axisType[i] )
		{
			case stvBlock:
			case stvCyclic:
			case stvExplicit:
				pindex++;
				break;

			case stvCollapse:
				break;
		}
	}

	/* Convert Index */

	switch ( D->axisType[ dindex ] )
	{
		case stvBlock:
		{
			index = gindex
				- ( F->paddr[ pindex ] * D->axisInfo[ dindex ] )
				- D->glb[ dindex ];

			break;
		}

		case stvCyclic:
		{
			cyclesize = D->axisInfo[ dindex ];

			size = cyclesize * D->pshape[ pindex ];

			gindex -= D->glb[ dindex ];

			index = ( ( gindex / size ) * cyclesize )
				+ ( gindex % size )
				- ( F->paddr[ pindex ] * cyclesize );

			break;
		}

		case stvExplicit:
		{
			index = gindex - D->axisInfo[ dindex ];

			break;
		}

		case stvCollapse:
		{
			index = gindex - D->glb[ dindex ];

			break;
		}
	}

	return( index );
}

