
/* $Id: stvfulib.c,v 1.5 1999/03/02 17:52:55 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided `as is'
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* stvfulib.m4 - macros to build FORTRAN interface for stv_ulib  */


#include <signal.h>
#include <stdio.h>

#include "pvm3.h"
#include "pvmwin.h"

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


void __stdcall STVFINIT (appname_ptr, appname_len, msgtag, nproc, nodeid, info)
char * appname_ptr; int appname_len;
int *msgtag, *nproc, *nodeid, *info;
{
	char cappname[ stvMaxNameLen + 1 ];

	/*
	 * Copy the appname name to make sure there's
	 * a NUL at the end.
	 */

	if ( ftocstr( cappname, sizeof(cappname), appname_ptr, 
		appname_len ) )
	{
		*info = stvStatusBadParam;
		return;
	}

	*info = stv_init( cappname, *msgtag, *nproc, *nodeid );

	return;
}


void __stdcall STVFDECOMPDEFINE (dataDim, axisType, axisInfo1, axisInfo2, glb, gub, prank, pshape, decompId)
int *dataDim, *axisType, *axisInfo1, *axisInfo2, *glb, *gub, *prank, *pshape, *decompId;
{
	int *axisTypeActual;
	int *axisInfo1Actual;
	int *axisInfo2Actual;
	int *pshapeActual;

	axisTypeActual = ( axisType[0] < 0 ) ? stvLocalArray : axisType;

	axisInfo1Actual = ( axisInfo1[0] < 0 ) ? stvLocalArray : axisInfo1;
	axisInfo2Actual = ( axisInfo2[0] < 0 ) ? stvLocalArray : axisInfo2;

	pshapeActual = ( pshape[0] < 0 ) ? stvLocalArray : pshape;

	*decompId = stv_decompDefine( *dataDim, axisTypeActual, 
		axisInfo1Actual, axisInfo2Actual, glb, gub,
		*prank, pshapeActual );

	return;
}


void __stdcall STVFFIELDDEFINE (var, name_ptr, name_len, decompID, arrayOffset, arrayDecl, type, paddr, aflag, fieldId)
char * name_ptr; int name_len;
STV_VALUE var;
int *decompID, *arrayOffset, *arrayDecl, *type, *paddr, *aflag,
	*fieldId;
{
	char cvarname[ stvMaxNameLen + 1 ];

	int *arrayOffsetActual;
	int *paddrActual;

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( ftocstr( cvarname, sizeof(cvarname), name_ptr, 
		name_len ) )
	{
		*fieldId = stvStatusBadParam;
		return;
	}

	arrayOffsetActual = ( arrayOffset[0] < 0 )
			? stvNoFieldOffsets : arrayOffset;

	paddrActual = ( paddr[0] < 0 ) ? stvLocalArray : paddr;

	*fieldId = stv_fieldDefine( var, cvarname, *decompID,
			arrayOffsetActual, arrayDecl, *type, paddrActual, *aflag );

	stv_setopt( *fieldId, stvOptFieldStorage, stvColumnMajor );

	return;
}


void __stdcall STVFPARAMDEFINE (var, name_ptr, name_len, type, aflag, paramId)
char * name_ptr; int name_len;
STV_VALUE var;
int  *type, *aflag, *paramId;
{
	char cvarname[ stvMaxNameLen + 1 ];

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( ftocstr( cvarname, sizeof(cvarname), name_ptr, 
		name_len ) )
	{
		*paramId = stvStatusBadParam;
		return;
	}

	*paramId = stv_paramDefine( cvarname, var, *type, *aflag );

	return;
}


void __stdcall STVFPARTICLEDEFINE (name_ptr, name_len, dataDim, glb, gub, nproc, get_particle, make_particle, particleId)
char * name_ptr; int name_len;
int *dataDim, *glb, *gub, *nproc, *particleId;
STV_GET_PARTICLE get_particle;
STV_MAKE_PARTICLE make_particle;
{
	char cvarname[ stvMaxNameLen + 1 ];

	char *tmp;

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( ftocstr( cvarname, sizeof(cvarname), name_ptr, 
		name_len ) )
	{
		*particleId = stvStatusBadParam;
		return;
	}

	*particleId = stv_particleDefine( cvarname, *dataDim, glb, gub, 
		*nproc, get_particle, make_particle );

	return;
}


void __stdcall STVFPARTICLEFIELDDEFINE (name_ptr, name_len, particleId, type, nelems, get_pfield, set_pfield, aflag, fieldId)
char * name_ptr; int name_len;
int *particleId, *type, *nelems;
STV_GET_PFIELD get_pfield;
STV_SET_PFIELD set_pfield;
int *aflag, *fieldId;
{
	char cvarname[ stvMaxNameLen + 1 ];

	char *tmp;

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( ftocstr( cvarname, sizeof(cvarname), name_ptr, 
		name_len ) )
	{
		*fieldId = stvStatusBadParam;
		return;
	}

	*fieldId = stv_particleFieldDefine( cvarname, *particleId, 
		*type, *nelems, get_pfield, set_pfield, *aflag );

	return;
}


void __stdcall STVFDUMPLOCALFIELDBOUNDS (fieldId, verbose, info)
int *fieldId;
int *verbose;
int *info;
{
	*info = stv_dumpLocalFieldBounds( *fieldId, *verbose );

	return;
}


void __stdcall STVFSENDTOFE (info)
int *info;
{
	*info = stv_sendToFE();

	return;
}


void __stdcall STVFSETOPT (what, option, value, oldvalue)
int *what;
int *option;
int *value;
int *oldvalue;
{
	*oldvalue = stv_setopt( *what, *option, *value );

	return;
}


void __stdcall STVFGETOPT (what, option, value)
int *what;
int *option;
int *value;
{
	*value = stv_getopt( *what, *option );

	return;
}


void __stdcall STVFTIMEOF (value, units, tvalue)
int *value;
int *units;
int *tvalue;
{
	*tvalue = stv_time_of( *value, *units );

	return;
}


void __stdcall STVFVALUEOFTIME (tvalue, value)
int *tvalue;
int *value;
{
	*value = stv_value_of_time( *tvalue );

	return;
}


void __stdcall STVFUNITOFTIME (tvalue, units)
int *tvalue;
int *units;
{
	*units = stv_unit_of_time( *tvalue );

	return;
}


void __stdcall STVFISTIMEVALUE (tvalue, info)
int *tvalue;
int *info;
{
	*info = stv_is_time_value( *tvalue );

	return;
}


void __stdcall STVFSLEEP (secs)
int *secs;
{
         Sleep( (long) *secs );

	return;
}


void __stdcall STVFSIGTRAP (info)
int *info;
{
	extern void stvsigquit(int sig);

	/* signal( SIGQUIT, stvsigquit ); */

	signal( SIGTERM, stvsigquit );

	*info = 0;

	return;
}


void 
stvsigquit(sig)
int sig;
{
	sprintf( STVERRTXT, "caught signal %d", sig );

	STV_ERROR( "stvsigquit", stvStatusBail, STVERRTXT );

	exit( sig );
}

void __stdcall STVFISPARAMCHANGED (paramId, info)
int *paramId;
int *info;
{
	*info = stv_isParamChanged(*paramId);

	return;
}


void __stdcall STVFVERSION (version_ptr, version_len)
char * version_ptr; int version_len;
{
	char *rcopy;

	rcopy = stv_version();

	ctofstr( version_ptr, version_len, rcopy );

	return;
}


void __stdcall STVFVERSIONNAME (name_ptr, name_len)
char * name_ptr; int name_len;
{
	char *rcopy;

	rcopy = stv_version_name();

	ctofstr( name_ptr, name_len, rcopy );

	return;
}

