
/* $Id: stvfulib.m4,v 3.16 2000/09/26 15:54:08 kohl Exp $ */


/*
 *                      CUMULVS version 1.0:
 *      Computational Steering and Interactive Visualization
 *                   in Distributed Applications
 *           Oak Ridge National Laboratory, Oak Ridge TN.
 *      Authors:  James Arthur Kohl and Philip M. Papadopoulos
 *                   (C) 1996 All Rights Reserved
 *
 *                              NOTICE
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose and without fee is hereby granted
 * provided that the above copyright notice appear in all copies and
 * that both the copyright notice and this permission notice appear
 * in supporting documentation.
 *
 * Neither the Institution, Oak Ridge National Laboratory, nor the
 * Authors make any representations about the suitability of this
 * software for any purpose.  This software is provided ``as is''
 * without express or implied warranty.
 *
 * CUMULVS was funded by the U.S. Department of Energy.
 */


/* stvfulib.m4 - macros to build FORTRAN interface for stv_ulib  */


#include <signal.h>
#include <stdio.h>

#include "stv.h"
#include "stvmac.h"
#include "stv_glob.h"


void
FUNCTION(stvfinit) ARGS(`STRING_ARG(appname), msgtag, ntasks, nodeid, info')
STRING_ARG_DECL(appname);
int *msgtag, *ntasks, *nodeid, *info;
{
	char cappname[ stvMaxNameLen + 1 ];

	/*
	 * Copy the appname name to make sure there's
	 * a NUL at the end.
	 */

	if ( stv_f_to_c_str( cappname, sizeof(cappname),
			STRING_PTR(appname), STRING_LEN(appname) ) )
	{
		*info = stvStatusBadParam;
		return;
	}

	*info = stv_init( cappname, *msgtag, *ntasks, *nodeid );

	return;
}


void
FUNCTION(stvfdecompdefine) ARGS(`dataDim, axisType, axisInfo1, axisInfo2, glb, gub, prank, pshape, decompId')
int *dataDim, *axisType, *axisInfo1, *axisInfo2, *glb, *gub, *prank, *pshape, *decompId;
{
	int *axisTypeActual;
	int *axisInfo1Actual;
	int *axisInfo2Actual;
	int *pshapeActual;

	axisTypeActual = ( axisType[0] < 0 ) ? stvLocalArray : axisType;

	axisInfo1Actual = ( axisInfo1[0] < 0 ) ? stvLocalArray : axisInfo1;
	axisInfo2Actual = ( axisInfo2[0] < 0 ) ? stvLocalArray : axisInfo2;

	pshapeActual = ( pshape[0] < 0 ) ? stvLocalArray : pshape;

	*decompId = stv_decompDefine( *dataDim, axisTypeActual, 
		axisInfo1Actual, axisInfo2Actual, glb, gub,
		*prank, pshapeActual );

	return;
}


void
FUNCTION(stvfdecompaddexplicitbounds) ARGS(`decompId, axis, lowerBound, upperBound, info')
int *decompId, *axis, *lowerBound, *upperBound, *info;
{
	*info = stv_decompAddExplicitBounds( *decompId, *axis, 
		*lowerBound, *upperBound );

	return;
}


void
FUNCTION(stvfdecompaddexplicitvbounds) ARGS(`decompId, axis, lowerBounds, upperBounds, nbounds, info')
int *decompId, *axis, *lowerBounds, *upperBounds, *nbounds, *info;
{
	*info = stv_decompAddExplicitVBounds( *decompId, *axis, 
		lowerBounds, upperBounds, *nbounds );

	return;
}


void
FUNCTION(stvffielddefine) ARGS(`var, STRING_ARG(name), decompID, arrayOffset, arrayDecl, type, paddr, aflag, fieldId')
STRING_ARG_DECL(name);
STV_VALUE var;
int *decompID, *arrayOffset, *arrayDecl, *type, *paddr, *aflag,
	*fieldId;
{
	char cvarname[ stvMaxNameLen + 1 ];

	int *arrayOffsetActual;
	int *paddrActual;

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( stv_f_to_c_str( cvarname, sizeof(cvarname),
			STRING_PTR(name), STRING_LEN(name) ) )
	{
		*fieldId = stvStatusBadParam;
		return;
	}

	arrayOffsetActual = ( arrayOffset[0] < 0 )
			? stvNoFieldOffsets : arrayOffset;

	paddrActual = ( paddr[0] < 0 ) ? stvLocalArray : paddr;

	*fieldId = stv_fieldDefine( var, cvarname, *decompID,
			arrayOffsetActual, arrayDecl, *type, paddrActual, *aflag );

	stv_setopt( *fieldId, stvOptFieldStorage, stvColumnMajor );

	return;
}


void
FUNCTION(stvfparamdefine) ARGS(`var, STRING_ARG(name), type, aflag, paramId')
STRING_ARG_DECL(name);
STV_VALUE var;
int  *type, *aflag, *paramId;
{
	char cvarname[ stvMaxNameLen + 1 ];

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( stv_f_to_c_str( cvarname, sizeof(cvarname),
			STRING_PTR(name), STRING_LEN(name) ) )
	{
		*paramId = stvStatusBadParam;
		return;
	}

	*paramId = stv_paramDefine( cvarname, var, *type, *aflag );

	return;
}


void
FUNCTION(stvfparticledefine) ARGS(`STRING_ARG(name), dataDim, glb, gub, ntasks, get_particle, make_particle, particleId')
STRING_ARG_DECL(name);
int *dataDim, *glb, *gub, *ntasks, *particleId;
STV_GET_PARTICLE get_particle;
STV_MAKE_PARTICLE make_particle;
{
	char cvarname[ stvMaxNameLen + 1 ];

	char *tmp;

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( stv_f_to_c_str( cvarname, sizeof(cvarname),
			STRING_PTR(name), STRING_LEN(name) ) )
	{
		*particleId = stvStatusBadParam;
		return;
	}

	*particleId = stv_particleDefine( cvarname, *dataDim, glb, gub, 
		*ntasks, get_particle, make_particle );

	return;
}


void
FUNCTION(stvfparticlefielddefine) ARGS(`STRING_ARG(name), particleId, type, nelems, get_pfield, set_pfield, aflag, fieldId')
STRING_ARG_DECL(name);
int *particleId, *type, *nelems;
STV_GET_PFIELD get_pfield;
STV_SET_PFIELD set_pfield;
int *aflag, *fieldId;
{
	char cvarname[ stvMaxNameLen + 1 ];

	char *tmp;

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( stv_f_to_c_str( cvarname, sizeof(cvarname),
			STRING_PTR(name), STRING_LEN(name) ) )
	{
		*fieldId = stvStatusBadParam;
		return;
	}

	*fieldId = stv_particleFieldDefine( cvarname, *particleId, 
		*type, *nelems, get_pfield, set_pfield, *aflag );

	return;
}


void
FUNCTION(stvfpfielddefine) ARGS(`STRING_ARG(name), particleId, type, nelems, get_pfield, get_pfield_arg, set_pfield, set_pfield_arg, aflag, fieldId')
STRING_ARG_DECL(name);
int *particleId, *type, *nelems;
STV_GET_PFIELD get_pfield;
STV_VALUE get_pfield_arg;
STV_SET_PFIELD set_pfield;
STV_VALUE set_pfield_arg;
int *aflag, *fieldId;
{
	char cvarname[ stvMaxNameLen + 1 ];

	char *tmp;

	/*
	 * Copy the name to make sure there's
	 * a NUL at the end.
	 */

	if ( stv_f_to_c_str( cvarname, sizeof(cvarname),
			STRING_PTR(name), STRING_LEN(name) ) )
	{
		*fieldId = stvStatusBadParam;
		return;
	}

	*fieldId = stv_pfieldDefine( cvarname,
		*particleId, *type, *nelems,
		get_pfield, get_pfield_arg, set_pfield, set_pfield_arg,
		*aflag );

	return;
}


void
FUNCTION(stvfdumplocalfieldbounds) ARGS(`fieldId, verbose, info')
int *fieldId;
int *verbose;
int *info;
{
	*info = stv_dumpLocalFieldBounds( *fieldId, *verbose );

	return;
}


void
FUNCTION(stvfsendtofe) ARGS(`info')
int *info;
{
	*info = stv_sendToFE();

	return;
}


void
FUNCTION(stvfsendreadydata) ARGS(`sendflag, info')
int *sendflag, *info;
{
	*info = stv_sendReadyData( *sendflag );

	return;
}


void
FUNCTION(stvfsetopt) ARGS(`what, option, value, oldvalue')
int *what;
int *option;
int *value;
int *oldvalue;
{
	*oldvalue = stv_setopt( *what, *option, *value );

	return;
}


void
FUNCTION(stvfgetopt) ARGS(`what, option, value')
int *what;
int *option;
int *value;
{
	*value = stv_getopt( *what, *option );

	return;
}


void
FUNCTION(stvftimeof) ARGS(`value, units, tvalue')
int *value;
int *units;
int *tvalue;
{
	*tvalue = stv_time_of( *value, *units );

	return;
}


void
FUNCTION(stvfvalueoftime) ARGS(`tvalue, value')
int *tvalue;
int *value;
{
	*value = stv_value_of_time( *tvalue );

	return;
}


void
FUNCTION(stvfunitoftime) ARGS(`tvalue, units')
int *tvalue;
int *units;
{
	*units = stv_unit_of_time( *tvalue );

	return;
}


void
FUNCTION(stvfistimevalue) ARGS(`tvalue, info')
int *tvalue;
int *info;
{
	*info = stv_is_time_value( *tvalue );

	return;
}


void
FUNCTION(stvfsleep) ARGS(`secs')
int *secs;
{
	stvSleep( (long) *secs );

	return;
}


void
FUNCTION(stvfsigtrap) ARGS(`info')
int *info;
{
	extern void stvsigquit();

	signal( SIGQUIT, stvsigquit );

	signal( SIGTERM, stvsigquit );

	*info = 0;

	return;
}


void
stvsigquit(sig)
int sig;
{
	sprintf( STVERRTXT, "caught signal %d", sig );

	STV_ERROR( "stvsigquit", stvStatusBail, STVERRTXT );

	exit( sig );
}

void
FUNCTION(stvfisparamchanged) ARGS(`paramId, info')
int *paramId;
int *info;
{
	*info = stv_isParamChanged(*paramId);

	return;
}


void
FUNCTION(stvfversion) ARGS(`STRING_ARG(version)')
STRING_ARG_DECL(version);
{
	char *rcopy;

	rcopy = stv_version();

	ctofstr( STRING_PTR(version), STRING_LEN(version), rcopy );

	return;
}


void
FUNCTION(stvfversionname) ARGS(`STRING_ARG(name)')
STRING_ARG_DECL(name);
{
	char *rcopy;

	rcopy = stv_version_name();

	ctofstr( STRING_PTR(name), STRING_LEN(name), rcopy );

	return;
}

